import ctypes
from functools import reduce
import numpy as np
import sys
from dataclasses import dataclass
from typing import Any, List, Tuple

from . import hat_file

# hat_declared_type : [ ctype, dtype_str ]
ARG_TYPES = {
    "int8_t*" : [ ctypes.c_int8, "int8" ],
    "int16_t*" : [ ctypes.c_int16, "int16" ],
    "int32_t*" : [ ctypes.c_int32, "int32" ],
    "int64_t*" : [ ctypes.c_int64, "int64" ],
    "uint8_t*" : [ ctypes.c_uint8, "uint8" ],
    "uint16_t*" : [ ctypes.c_uint16, "uint16" ],
    "uint32_t*" : [ ctypes.c_uint32, "uint32" ],
    "uint64_t*" : [ ctypes.c_uint64, "uint64" ],
    "float16_t*" : [ ctypes.c_uint16, "float16" ], # same bitwidth as uint16
    "bfloat16_t*" : [ ctypes.c_uint16, "bfloat16" ],
    "float*" : [ ctypes.c_float, "float32" ],
    "double*" : [ ctypes.c_double, "float64" ],
}
CTYPE_ENTRY = 0
DTYPE_ENTRY = 1

@dataclass
class ArgInfo:
    """Extracts necessary information from the description of a function argument in a hat file"""
    hat_declared_type: str
    numpy_shape: Tuple[int, ...]
    numpy_strides: Tuple[int, ...]
    numpy_dtype: type
    element_num_bytes: int
    element_strides: Tuple[int, ...]
    total_element_count: int
    total_byte_size: int
    ctypes_pointer_type: Any
    usage: hat_file.UsageType = None

    def _get_type(self, type_str):
        if type_str == "bfloat16":
            from bfloat16 import bfloat16
            return bfloat16

        return np.dtype(type_str)

    def __init__(self, param_description: hat_file.Parameter):
        self.hat_declared_type = param_description.declared_type
        self.numpy_shape = tuple(param_description.shape)
        self.usage = param_description.usage

        if not self.hat_declared_type in ARG_TYPES:
            raise NotImplementedError(f"Unsupported declared_type {self.hat_declared_type} in hat file")

        self.ctypes_pointer_type = ctypes.POINTER(ARG_TYPES[self.hat_declared_type][CTYPE_ENTRY])
        dtype_entry = ARG_TYPES[self.hat_declared_type][DTYPE_ENTRY]
        self.numpy_dtype = self._get_type(dtype_entry)
        self.element_num_bytes = 2 if dtype_entry == "bfloat16" else self.numpy_dtype.itemsize
        self.element_strides = param_description.affine_map
        self.numpy_strides = tuple([self.element_num_bytes * x for x in self.element_strides])

        major_dim = self.element_strides.index(max(self.element_strides))
        self.total_element_count = self.numpy_shape[major_dim] * self.element_strides[major_dim]
        self.total_byte_size = self.element_num_bytes * self.total_element_count


# TODO: Update this to take a HATFunction instead, instead of arg_infos and function_name
def verify_args(args: List, arg_infos: List[ArgInfo], function_name: str):
    """ Verifies that a list of arguments matches a list of argument descriptions in a HAT file
    """
    # check number of args
    if len(args) != len(arg_infos):
        sys.exit(f"Error calling {function_name}(...): expected {len(arg_infos)} arguments but received {len(args)}")

    # for each arg
    for i in range(len(args)):
        arg = args[i]
        arg_info = arg_infos[i]

        # confirm that the arg is a numpy ndarray
        if not isinstance(arg, np.ndarray):
            sys.exit(
                "Error calling {function_name}(...): expected argument {i} to be <class 'numpy.ndarray'> but received {type(arg)}"
            )

        # confirm that the arg dtype matches the dexcription in the hat package
        if arg_info.numpy_dtype != arg.dtype:
            sys.exit(
                f"Error calling {function_name}(...): expected argument {i} to have dtype={arg_info.numpy_dtype} but received dtype={arg.dtype}"
            )

        # confirm that the arg shape is correct
        if arg_info.numpy_shape != arg.shape:
            sys.exit(
                f"Error calling {function_name}(...): expected argument {i} to have shape={arg_info.numpy_shape} but received shape={arg.shape}"
            )

        # confirm that the arg strides are correct
        if arg_info.numpy_strides != arg.strides:
            sys.exit(
                f"Error calling {function_name}(...): expected argument {i} to have strides={arg_info.numpy_strides} but received strides={arg.strides}"
            )



#!/usr/bin/env python3

import io
import os
import platform
import shlex
import shutil
import subprocess
import sys

from .hat_file import OperatingSystem


def _preprocess_command(command_to_run, shell):
    if platform.system() == "Windows":
        return command_to_run
    elif type(command_to_run) == str and not shell:
        return shlex.split(command_to_run)
    elif type(command_to_run) == list and shell:
        return subprocess.list2cmdline(command_to_run)
    else:
        return command_to_run


def _dump_file_contents(iostream):
    if isinstance(iostream, io.TextIOBase):
        with open(iostream.name, "r") as f:
            print(f.name)
            print(f.read())


def run_command(
    command_to_run, working_directory=None, stdout=None, stderr=None, shell=False, pretend=False, quiet=True
):
    if not working_directory:
        working_directory = os.getcwd()

    if not quiet:
        print(f"\ncd {working_directory}")
        print(f"{command_to_run}\n")

    command_to_run = _preprocess_command(command_to_run, shell)

    if not pretend:
        with subprocess.Popen(command_to_run, close_fds=(platform.system() != "Windows"), shell=shell, stdout=stdout,
                              stderr=stderr, cwd=working_directory) as proc:

            proc.wait()
            if proc.returncode:
                _dump_file_contents(stderr)
                _dump_file_contents(stdout)
                raise subprocess.CalledProcessError(proc.returncode, command_to_run)


def get_platform():
    """Returns the current platform: Linux, Windows, or OS X"""

    if sys.platform.startswith("linux"):
        return OperatingSystem.Linux
    if sys.platform == "win32":
        return OperatingSystem.Windows
    if sys.platform == "darwin":
        return OperatingSystem.MacOS

    sys.exit(f"ERROR: Unsupported operating system: {sys.platform}")


def linux_ensure_compiler_in_path():
    """Ensures that PATH contains the $CXX/$CC (or gcc) compiler.
    Prompts the user if not found."""
    compiler = os.environ.get("CXX") or (os.environ.get("CC") or "gcc")
    if not shutil.which(compiler):
        sys.exit('ERROR: Could not find any valid C or C++ compiler, please install gcc before continuing')


def windows_ensure_compiler_in_path():
    """Ensures that PATH contains the cl.exe compiler from Microsoft Visual Studio.
    Prompts the user if not found."""
    import vswhere
    vs_path = vswhere.get_latest_path()
    if not vs_path:
        sys.exit("ERROR: Could not find Visual Studio, please ensure that you have Visual Studio installed")

    # Check if cl.exe is in PATH
    if not shutil.which("cl"):    # returns 0 if found, !0 otherwise
        vcvars_script_path = os.path.join(vs_path, r"VC\Auxiliary\Build\vcvars64.bat")
        sys.exit(
            f'ERROR: Could not find cl.exe, please run "{vcvars_script_path}" (including quotes) to setup your command prompt'
        )


def ensure_compiler_in_path():
    platform = get_platform()
    if platform == OperatingSystem.Windows:
        windows_ensure_compiler_in_path()
    else:
        linux_ensure_compiler_in_path()



import ctypes
import pathlib
import numpy as np
from typing import List

from .arg_info import ArgInfo, verify_args
from .callable_func import CallableFunc
from .hat_file import Function
from .gpu_headers import ROCM_HEADER_MAP
from .pyhip.hip import *
from .pyhip.hiprtc import *


def initialize_rocm():
    # Initialize ROCM Driver API
    hipInit(0)


def compile_rocm_program(rocm_src_path: pathlib.Path, func_name):
    src = rocm_src_path.read_text()

    prog = hiprtcCreateProgram(
        source=src,
        name=func_name + ".cu",
        header_names=ROCM_HEADER_MAP.keys(),
        header_sources=ROCM_HEADER_MAP.values()
    )
    device_properties = hipGetDeviceProperties(0)
    hiprtcCompileProgram(prog, [f'--offload-arch={device_properties.gcnArchName}', '-D__HIP_PLATFORM_AMD__'])
    code = hiprtcGetCode(prog)

    return code


def get_func_from_rocm_program(rocm_program, func_name):
    rocm_module = hipModuleLoadData(rocm_program)
    kernel = hipModuleGetFunction(rocm_module, func_name)
    return kernel

cached_mem=[]

def allocate_rocm_mem(benchmark: bool, arg_infos: List[ArgInfo], gpu_id: int):
    device_mem = []
    for arg in arg_infos:
        try:
            memory_cache = cached_mem[gpu_id]
            if benchmark and arg.total_byte_size in memory_cache:
                mem = memory_cache[arg.total_byte_size]
            else:
                mem = hipMalloc(arg.total_byte_size)
                if benchmark:
                    memory_cache[arg.total_byte_size] = mem
        except:
            free_rocm_mem(device_mem)
            raise
        device_mem.append(mem)

    return device_mem


def free_rocm_mem(args):
    for arg in args:
        hipFree(arg)


def transfer_mem_host_to_rocm(device_args: List, host_args: List[np.array], arg_infos: List[ArgInfo]):
    for device_arg, host_arg, arg_info in zip(device_args, host_args, arg_infos):
        if 'input' in arg_info.usage.value:
            hipMemcpy_htod(dst=device_arg, src=host_arg.ctypes.data, count=arg_info.total_byte_size)


def transfer_mem_rocm_to_host(device_args: List, host_args: List[np.array], arg_infos: List[ArgInfo]):
    for device_arg, host_arg, arg_info in zip(device_args, host_args, arg_infos):
        if 'output' in arg_info.usage.value:
            hipMemcpy_dtoh(dst=host_arg.ctypes.data, src=device_arg, count=arg_info.total_byte_size)


def device_args_to_ptr_list(device_args: List):
    ptrs = [np.array([int(d_arg)], dtype=np.uint64) for d_arg in device_args]
    ptrs = np.array([ptr.ctypes.data for ptr in ptrs], dtype=np.uint64)

    return ptrs


_HSACO_CACHE = {}


class RocmCallableFunc(CallableFunc):

    def __init__(self, func: Function, rocm_src_path: str) -> None:
        super().__init__()
        self.hat_func = func
        self.func_name = func.name
        self.kernel = None
        hat_arg_descriptions = func.arguments
        self.arg_infos = [ArgInfo(d) for d in hat_arg_descriptions]
        self.launch_params = func.launch_parameters
        self.device_mem = None
        self.ptrs = None
        self.stream = None
        self.start_event = None
        self.stop_event = None
        self.exec_time = 0.
        self.rocm_src_path = rocm_src_path

    def init_runtime(self, benchmark: bool, gpu_id: int):
        if not benchmark:
            initialize_rocm()

        hipSetDevice(gpu_id)

        # Add a separate cache for each gpu since device memory is not shareable (duh!)
        while len(cached_mem) <= gpu_id:
            cached_mem.append({})

        rocm_program = _HSACO_CACHE.get(self.rocm_src_path)
        if not rocm_program:
            _HSACO_CACHE[self.rocm_src_path] = rocm_program = compile_rocm_program(self.rocm_src_path, self.func_name)

        self.kernel = get_func_from_rocm_program(rocm_program, self.func_name)

    def cleanup_runtime(self, benchmark: bool):
        pass

    def init_main(self, benchmark: bool, warmup_iters=0, args=[], gpu_id: int=0):
        verify_args(args, self.arg_infos, self.func_name)
        self.device_mem = allocate_rocm_mem(benchmark, self.arg_infos, gpu_id)

        if not benchmark:
            transfer_mem_host_to_rocm(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)

        class DataStruct(ctypes.Structure):
            _fields_ = [(f"arg{i}", ctypes.c_void_p) for i in range(len(self.arg_infos))]

        self.data = DataStruct(*self.device_mem)

        self.start_event = hipEventCreate()
        self.stop_event = hipEventCreate()

        for _ in range(warmup_iters):
            hipModuleLaunchKernel(
                self.kernel,
                *self.launch_params,    # [ grid[x-z], block[x-z] ]
                0,    # dynamic shared memory
                0,    # stream
                self.data,    # data
            )

    def main(self, benchmark: bool, iters=1, batch_size=1, args=[]) -> float:
        batch_timings: List[float] = []
        for _ in range(batch_size):
            hipEventRecord(self.start_event)

            for _ in range(iters):
                hipModuleLaunchKernel(
                    self.kernel,
                    *self.launch_params,    # [ grid[x-z], block[x-z] ]
                    0,    # dynamic shared memory
                    0,    # stream
                    self.data,    # data
                )

            hipEventRecord(self.stop_event)
            hipEventSynchronize(self.stop_event)
            batch_time = hipEventElapsedTime(self.start_event, self.stop_event)
            batch_timings.append(batch_time)
            self.exec_time += batch_time

            if not benchmark:
                hipDeviceSynchronize()

        self.exec_time /= (iters * batch_size)
        return batch_timings

    def cleanup_main(self, benchmark: bool, args=[]):
        # If there's no device mem, that means allocation during initialization failed, which means nothing else needs to be cleaned up either
        if not benchmark and self.device_mem:
            transfer_mem_rocm_to_host(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)
            free_rocm_mem(self.device_mem)
        hipDeviceSynchronize()

        if self.start_event:
            hipEventDestroy(self.start_event)

        if self.stop_event:
            hipEventDestroy(self.stop_event)


def create_loader_for_device_function(device_func: Function, hat_dir_path: str):
    if not device_func.provider:
        raise RuntimeError("Expected a provider for the device function")

    rocm_src_path: pathlib.Path = pathlib.Path(hat_dir_path) / device_func.provider

    return RocmCallableFunc(device_func, rocm_src_path)



#!/usr/bin/env python3
"""Converts a statically-linked HAT package into a Dynamically-linked HAT package

HAT packages come in two varieties: statically-linked and dynamically-linked. A statically-linked
HAT package contains a binary file with the extension '.o', '.obj', '.a', or `.lib`. A dynamically-linked
HAT package contains a binary file with the extension '.dll' or '.so'. This tool converts a
statically-linked HAT package into a dynamically-linked HAT package.

To use the tool, point it to the '.hat' file associated with the statically-linked package (the
'.hat' file knows where to find the associated binary file), and provide a filename for the new
'.hat' file.

Dependencies on Windows:
* the cl.exe command-line compiler, available with Microsoft Visual Studio

Dependencies on Linux / macOS:
* the gcc command-line compiler
"""

import sys
import os
import argparse
import shutil
from secrets import token_hex

from .hat_file import HATFile, OperatingSystem
from .hat_package import HATPackage
from .platform_utilities import get_platform, ensure_compiler_in_path, run_command


def linux_create_dynamic_package(input_hat_path,
                                 input_hat_binary_path,
                                 output_hat_path,
                                 hat_file,
                                 quiet=True):
    """Creates a dynamic HAT (.so) from a static HAT (.o/.a) on a Linux/macOS platform"""
    # Confirm that this is a static hat library
    _, extension = os.path.splitext(input_hat_binary_path)
    if extension not in [".o", ".a"]:
        sys.exit(
            f"ERROR: Expected input library to have extension .o or .a, but received {input_hat_binary_path} instead"
        )

    # Create a C source file to resolve inline functions defined in the static HAT package
    include_path = os.path.dirname(input_hat_binary_path)
    inline_c_path = os.path.join(include_path, "inline.c")
    inline_obj_path = os.path.join(include_path, "inline.o")
    with open(inline_c_path, "w") as f:
        f.write(f"#include <{os.path.basename(input_hat_path)}>")
    # compile it separately so that we can suppress the warnings about the missing terminating ' character
    run_command(
        f'gcc -c -w -fPIC -o "{inline_obj_path}" -I"{include_path}" "{inline_c_path}"',
        quiet=quiet)

    # create new HAT binary
    prefix, _ = os.path.splitext(output_hat_path)
    # always create a new so file (avoids cases where so file is already loaded)
    suffix = token_hex(4)
    output_hat_binary_path = f"{prefix}_{suffix}.so"
    libraries = " ".join(
        [d.target_file for d in hat_file.dependencies.dynamic])
    run_command(
        f'gcc -shared -fPIC -o "{output_hat_binary_path}" "{inline_obj_path}" "{input_hat_binary_path}" {libraries}',
        quiet=quiet)

    # create new HAT file
    # previous dependencies are now part of the binary
    hat_file.dependencies.dynamic = []
    hat_file.dependencies.link_target = os.path.basename(
        output_hat_binary_path)
    hat_file.Serialize(output_hat_path)

    return HATPackage(output_hat_path)


def windows_create_dynamic_package(input_hat_path,
                                   input_hat_binary_path,
                                   output_hat_path,
                                   hat_file,
                                   quiet=True):
    """Creates a Windows dynamic HAT package (.dll) from a static HAT package (.obj/.lib)"""

    # Confirm that this is a static hat library
    _, extension = os.path.splitext(input_hat_binary_path)
    if extension not in [".obj", ".lib"]:
        sys.exit(
            f"ERROR: Expected input library to have extension .obj or .lib, but received {input_hat_binary_path} instead"
        )

    # Create all file in a directory named build
    if not os.path.exists("build"):
        os.mkdir("build")

    cwd = os.getcwd()
    try:
        os.chdir("build")

        # Create a C source file for the DLL entry point and compile in into an obj
        with open("dllmain.cpp", "w") as f:
            f.write("#include <windows.h>\n")
            # Resolve inline functions defined in the static HAT package
            f.write("#include <{}>\n".format(os.path.basename(input_hat_path)))
            f.write(
                "BOOL APIENTRY DllMain(HMODULE, DWORD, LPVOID) { return TRUE; }\n"
            )
        run_command(
            f'cl.exe /nologo /I"{os.path.dirname(input_hat_path)}" /Fodllmain.obj /c dllmain.cpp',
            quiet=quiet)

        # create the new HAT binary dll
        # always create a new dll (avoids case where dll is already loaded)
        suffix = token_hex(4)
        prefix, _ = os.path.splitext(output_hat_path)
        output_hat_binary_path = f"{prefix}_{suffix}.dll"

        function_descriptions = hat_file.functions
        function_names = [f.name for f in function_descriptions]
        exports = " /EXPORT:".join(function_names)

        libraries = " ".join(
            [d.target_file for d in hat_file.dependencies.dynamic])
        linker_command_line = f'link.exe /NOLOGO /DLL /FORCE:MULTIPLE /EXPORT:{exports} /OUT:out.dll dllmain.obj "{input_hat_binary_path}" {libraries}'
        run_command(linker_command_line, quiet=quiet)
        shutil.copyfile("out.dll", output_hat_binary_path)

        # create new HAT file
        # previous dependencies are now part of the binary
        hat_file.dependencies.dynamic = []
        hat_file.dependencies.link_target = os.path.basename(
            output_hat_binary_path)
        hat_file.Serialize("out.hat")
        shutil.copyfile("out.hat", output_hat_path)
    finally:
        os.chdir(cwd)  # restore the current working directory

    return HATPackage(output_hat_path)


def parse_args():
    """Parses and checks the command line arguments"""
    parser = argparse.ArgumentParser(
        description=
        "Creates a dynamically-linked HAT package from a statically-linked HAT package.\n"
        "Example:\n"
        "    hatlib.hat_to_dynamic input.hat output.hat\n")

    parser.add_argument(
        "input_hat_path",
        type=str,
        help=
        "Path to the existing HAT file, which represents a statically-linked HAT package"
    )
    parser.add_argument(
        "output_hat_path",
        type=str,
        help=
        "Path to the new HAT file, which will represent a dynamically-linked HAT package"
    )
    parser.add_argument('-v',
                        "--verbose",
                        action='store_true',
                        help="Enable verbose output")
    args = parser.parse_args()

    # check args
    if not os.path.exists(args.input_hat_path):
        sys.exit(f"ERROR: File {args.input_hat_path} not found")

    if os.path.abspath(args.input_hat_path) == os.path.abspath(
            args.output_hat_path):
        sys.exit("ERROR: Output file must be different from input file")

    _, extension = os.path.splitext(args.input_hat_path)
    if extension != ".hat":
        sys.exit(
            f"ERROR: Expected input file to have extension .hat, but received {extension} instead"
        )

    _, extension = os.path.splitext(args.output_hat_path)
    if extension != ".hat":
        sys.exit(
            f"ERROR: Expected output file to have extension .hat, but received {extension} instead"
        )

    return args


def create_dynamic_package(input_hat_path, output_hat_path, quiet=True):
    platform = get_platform()
    ensure_compiler_in_path()

    # load the function decscriptions and the library path from the hat file
    input_hat_path = os.path.abspath(input_hat_path)
    hat_file = HATFile.Deserialize(input_hat_path)

    # get the absolute path to the input binary
    input_hat_binary_filename = hat_file.dependencies.link_target
    input_hat_binary_path = os.path.join(os.path.dirname(input_hat_path),
                                         input_hat_binary_filename)

    # create the dynamic package
    output_hat_path = os.path.abspath(output_hat_path)
    if platform == OperatingSystem.Windows:
        windows_create_dynamic_package(input_hat_path,
                                       input_hat_binary_path,
                                       output_hat_path,
                                       hat_file,
                                       quiet=quiet)
    elif platform in [OperatingSystem.Linux, OperatingSystem.MacOS]:
        linux_create_dynamic_package(input_hat_path,
                                     input_hat_binary_path,
                                     output_hat_path,
                                     hat_file,
                                     quiet=quiet)


def main():
    args = parse_args()
    create_dynamic_package(args.input_hat_path,
                           args.output_hat_path,
                           quiet=not args.verbose)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Utility to parse the TOML metadata from HAT files
import os
import tomlkit
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Dict, List

# TODO : type-checking on leaf node values


def _read_toml_file(filepath):
    path = os.path.abspath(filepath)
    toml_doc = None
    with open(path, "r") as f:
        file_contents = f.read()
        toml_doc = tomlkit.parse(file_contents)
    return toml_doc


def _check_required_table_entry(table, key):
    if key not in table:
        # TODO : add more context to this error message
        raise ValueError(f"Invalid HAT file: missing required key {key}")


def _check_required_table_entries(table, keys):
    for key in keys:
        _check_required_table_entry(table, key)


class ParameterType(Enum):
    AffineArray = "affine_array"
    RuntimeArray = "runtime_array"
    Element = "element"
    Void = "void"


class UsageType(Enum):
    Input = "input"
    Output = "output"
    InputOutput = "input_output"


class CallingConventionType(Enum):
    StdCall = "stdcall"
    CDecl = "cdecl"
    FastCall = "fastcall"
    VectorCall = "vectorcall"
    Device = "devicecall"


class TargetType(Enum):
    CPU = "CPU"
    GPU = "GPU"


class OperatingSystem(Enum):
    Windows = "windows"
    MacOS = "macos"
    Linux = "linux"


@dataclass
class AuxiliarySupportedTable:
    AuxiliaryKey = "auxiliary"
    auxiliary: dict = field(default_factory=dict)

    def add_auxiliary_table(self, table):
        if len(self.auxiliary) > 0:
            table.add(self.AuxiliaryKey, self.auxiliary)

    @staticmethod
    def parse_auxiliary(table):
        if AuxiliarySupportedTable.AuxiliaryKey in table:
            return table[AuxiliarySupportedTable.AuxiliaryKey]
        else:
            return {}


@dataclass
class Description(AuxiliarySupportedTable):
    TableName: str = "description"
    comment: str = ""
    author: str = ""
    version: str = ""
    license_url: str = ""

    def to_table(self):
        description_table = tomlkit.table()
        description_table.add("comment", self.comment)
        description_table.add("author", self.author)
        description_table.add("version", self.version)
        description_table.add("license_url", self.license_url)

        self.add_auxiliary_table(description_table)

        return description_table

    @staticmethod
    def parse_from_table(table):
        return Description(
            author=table["author"],
            version=table["version"],
            license_url=table["license_url"],
            auxiliary=AuxiliarySupportedTable.parse_auxiliary(table)
        )


@dataclass
class Parameter:
    # All parameter keys
    name: str = ""
    description: str = ""
    logical_type: ParameterType = None
    declared_type: str = ""
    element_type: str = ""
    usage: UsageType = None

    # Affine array parameter keys
    shape: str = ""
    affine_map: list = field(default_factory=list)
    affine_offset: int = -1

    # Runtime array parameter keys
    size: str = ""

    def to_table(self):
        table = tomlkit.inline_table()
        table.append("name", self.name)
        table.append("description", self.description)
        table.append("logical_type", self.logical_type.value)
        table.append("declared_type", self.declared_type)
        table.append("element_type", self.element_type)
        table.append("usage", self.usage.value)

        if self.logical_type == ParameterType.AffineArray:
            table.append("shape", self.shape)
            table.append("affine_map", self.affine_map)
            table.append("affine_offset", self.affine_offset)
        elif self.logical_type == ParameterType.RuntimeArray:
            table.append("size", self.size)

        return table

    # TODO : change "usage" to "role" in schema
    @staticmethod
    def parse_from_table(param_table):
        required_table_entries = ["name", "description", "logical_type", "declared_type", "element_type", "usage"]
        _check_required_table_entries(param_table, required_table_entries)
        affine_array_required_table_entries = ["shape", "affine_map", "affine_offset"]
        runtime_array_required_table_entries = ["size"]

        name = param_table["name"]
        description = param_table["description"]
        logical_type = ParameterType(param_table["logical_type"])
        declared_type = param_table["declared_type"]
        element_type = param_table["element_type"]
        usage = UsageType(param_table["usage"])

        param = Parameter(
            name=name,
            description=description,
            logical_type=logical_type,
            declared_type=declared_type,
            element_type=element_type,
            usage=usage
        )

        if logical_type == ParameterType.AffineArray:
            _check_required_table_entries(param_table, affine_array_required_table_entries)
            param.shape = param_table["shape"]
            param.affine_map = param_table["affine_map"]
            param.affine_offset = param_table["affine_offset"]

        elif logical_type == ParameterType.RuntimeArray:
            _check_required_table_entries(param_table, runtime_array_required_table_entries)
            param.size = param_table["size"]

        return param


@dataclass
class Function(AuxiliarySupportedTable):
    # required
    arguments: List[Parameter] = field(default_factory=list)
    calling_convention: CallingConventionType = None
    description: str = ""
    hat_file: any = None
    link_target: Path = None
    name: str = ""
    return_info: Parameter = None

    # optional
    launch_parameters: list = field(default_factory=list)
    launches: str = ""
    provider: str = ""
    runtime: str = ""

    def to_table(self):
        table = tomlkit.table()
        table.add("name", self.name)
        table.add("description", self.description)
        table.add("calling_convention", self.calling_convention.value)
        arg_tables = [arg.to_table() for arg in self.arguments]
        arg_array = tomlkit.array()
        for arg_table in arg_tables:
            arg_array.append(arg_table)
        table.add(
            "arguments", arg_array
        )    # TODO : figure out why this isn't indenting after serialization in some cases

        if self.launch_parameters:
            table.add("launch_parameters", self.launch_parameters)

        if self.launches:
            table.add("launches", self.launches)

        if self.provider:
            table.add("provider", self.provider)

        if self.runtime:
            table.add("runtime", self.runtime)

        table.add("return", self.return_info.to_table())

        self.add_auxiliary_table(table)

        return table

    @staticmethod
    def parse_from_table(function_table):
        required_table_entries = ["name", "description", "calling_convention", "arguments", "return"]
        _check_required_table_entries(function_table, required_table_entries)
        arguments = [Parameter.parse_from_table(param_table) for param_table in function_table["arguments"]]

        launch_parameters = function_table["launch_parameters"] if "launch_parameters" in function_table else []

        launches = function_table["launches"] if "launches" in function_table else ""

        provider = function_table["provider"] if "provider" in function_table else ""

        runtime = function_table["runtime"] if "runtime" in function_table else ""

        return_info = Parameter.parse_from_table(function_table["return"])

        return Function(
            name=function_table["name"],
            description=function_table["description"],
            calling_convention=CallingConventionType(function_table["calling_convention"]),
            arguments=arguments,
            return_info=return_info,
            launch_parameters=launch_parameters,
            launches=launches,
            provider=provider,
            runtime=runtime,
            auxiliary=AuxiliarySupportedTable.parse_auxiliary(function_table)
        )


class FunctionTableCommon:

    def __init__(self, function_map):
        self.function_map = function_map
        self.functions = self.function_map.values()

    def to_table(self):
        func_table = tomlkit.table()
        for function_key in self.function_map:
            func_table.add(function_key, self.function_map[function_key].to_table())
        return func_table

    @classmethod
    def parse_from_table(cls, all_functions_table):
        function_map = {
            function_key: Function.parse_from_table(all_functions_table[function_key])
            for function_key in all_functions_table
        }
        return cls(function_map)


class FunctionTable(FunctionTableCommon):
    TableName = "functions"


class DeviceFunctionTable(FunctionTableCommon):
    TableName = "device_functions"


@dataclass
class Target:

    @dataclass
    class Required:

        @dataclass
        class CPU:
            TableName = TargetType.CPU.value

            # required
            architecture: str = ""
            extensions: list = field(default_factory=list)

            # optional
            runtime: str = ""

            def to_table(self):
                table = tomlkit.table()
                table.add("architecture", self.architecture)
                table.add("extensions", self.extensions)

                if self.runtime:
                    table.add("runtime", self.runtime)

                return table

            @staticmethod
            def parse_from_table(table):
                required_table_entries = ["architecture", "extensions"]
                _check_required_table_entries(table, required_table_entries)

                runtime = table.get("runtime", "")

                return Target.Required.CPU(
                    architecture=table["architecture"], extensions=table["extensions"], runtime=runtime
                )

        @dataclass
        class GPU:
            TableName = TargetType.GPU.value
            blocks: int = 0
            instruction_set_version: str = ""
            min_threads: int = 0
            min_global_memory_KB: int = 0
            min_shared_memory_KB: int = 0
            min_texture_memory_KB: int = 0
            model: str = ""
            runtime: str = ""

            def to_table(self):
                table = tomlkit.table()
                table.add("model", self.model)
                table.add("runtime", self.runtime)
                table.add("blocks", self.blocks)
                table.add("instruction_set_version", self.instruction_set_version)
                table.add("min_threads", self.min_threads)
                table.add("min_global_memory_KB", self.min_global_memory_KB)
                table.add("min_shared_memory_KB", self.min_shared_memory_KB)
                table.add("min_texture_memory_KB", self.min_texture_memory_KB)

                return table

            @staticmethod
            def parse_from_table(table):
                required_table_entries = [
                    "runtime",
                    "model",
                ]
                _check_required_table_entries(table, required_table_entries)

                return Target.Required.GPU(
                    runtime=table["runtime"],
                    model=table["model"],
                    blocks=table["blocks"],
                    instruction_set_version=table["instruction_set_version"],
                    min_threads=table["min_threads"],
                    min_global_memory_KB=table["min_global_memory_KB"],
                    min_shared_memory_KB=table["min_shared_memory_KB"],
                    min_texture_memory_KB=table["min_texture_memory_KB"]
                )

        TableName = "required"
        os: OperatingSystem = None
        cpu: CPU = None
        gpu: GPU = None

        def to_table(self):
            table = tomlkit.table()
            table.add("os", self.os.value)
            table.add(Target.Required.CPU.TableName, self.cpu.to_table())
            if self.gpu and self.gpu.runtime:
                table.add(Target.Required.GPU.TableName, self.gpu.to_table())
            return table

        @staticmethod
        def parse_from_table(table):
            required_table_entries = ["os", Target.Required.CPU.TableName]
            _check_required_table_entries(table, required_table_entries)
            cpu_info = Target.Required.CPU.parse_from_table(table[Target.Required.CPU.TableName])
            if Target.Required.GPU.TableName in table:
                gpu_info = Target.Required.GPU.parse_from_table(table[Target.Required.GPU.TableName])
            else:
                gpu_info = Target.Required.GPU()
            return Target.Required(os=table["os"], cpu=cpu_info, gpu=gpu_info)

    # TODO : support optimized_for table
    class OptimizedFor:
        TableName = "optimized_for"

        def to_table(self):
            return tomlkit.table()

        @staticmethod
        def parse_from_table(table):
            pass

    TableName = "target"
    required: Required = None
    optimized_for: OptimizedFor = None

    def to_table(self):
        table = tomlkit.table()
        table.add(Target.Required.TableName, self.required.to_table())
        if self.optimized_for is not None:
            table.add(Target.OptimizedFor.TableName, self.optimized_for.to_table())
        return table

    @staticmethod
    def parse_from_table(target_table):
        required_table_entries = [Target.Required.TableName]
        _check_required_table_entries(target_table, required_table_entries)
        required_data = Target.Required.parse_from_table(target_table[Target.Required.TableName])
        if Target.OptimizedFor.TableName in target_table:
            optimized_for_data = Target.OptimizedFor.parse_from_table(target_table[Target.OptimizedFor.TableName])
        else:
            optimized_for_data = Target.OptimizedFor()
        return Target(required=required_data, optimized_for=optimized_for_data)


@dataclass
class LibraryReference:
    name: str = ""
    version: str = ""
    target_file: str = ""

    def to_table(self):
        table = tomlkit.inline_table()
        table.append("name", self.name)
        table.append("version", self.version)
        table.append("target_file", self.target_file)
        return table

    @staticmethod
    def parse_from_table(table):
        return LibraryReference(name=table["name"], version=table["version"], target_file=table["target_file"])


@dataclass
class Dependencies(AuxiliarySupportedTable):
    TableName = "dependencies"
    link_target: str = ""
    deploy_files: list = field(default_factory=list)
    dynamic: list = field(default_factory=list)

    def to_table(self):
        table = tomlkit.table()
        table.add("link_target", self.link_target)
        table.add("deploy_files", self.deploy_files)

        dynamic_arr = tomlkit.array()
        for elt in self.dynamic:
            dynamic_arr.append(elt.to_table())
        table.add("dynamic", dynamic_arr)

        self.add_auxiliary_table(table)
        return table

    @staticmethod
    def parse_from_table(dependencies_table):
        required_table_entries = ["link_target", "deploy_files", "dynamic"]
        _check_required_table_entries(dependencies_table, required_table_entries)
        dynamic = [LibraryReference.parse_from_table(lib_ref_table) for lib_ref_table in dependencies_table["dynamic"]]
        return Dependencies(
            link_target=dependencies_table["link_target"],
            deploy_files=dependencies_table["deploy_files"],
            dynamic=dynamic,
            auxiliary=AuxiliarySupportedTable.parse_auxiliary(dependencies_table)
        )


@dataclass
class CompiledWith:
    TableName = "compiled_with"
    compiler: str = ""
    flags: str = ""
    crt: str = ""
    libraries: list = field(default_factory=list)

    def to_table(self):
        table = tomlkit.table()
        table.add("compiler", self.compiler)
        table.add("flags", self.flags)
        table.add("crt", self.crt)

        libraries_arr = tomlkit.array()
        for elt in self.libraries:
            libraries_arr.append(elt.to_table())
        table.add("libraries", libraries_arr)

        return table

    @staticmethod
    def parse_from_table(compiled_with_table):
        required_table_entries = ["compiler", "flags", "crt", "libraries"]
        _check_required_table_entries(compiled_with_table, required_table_entries)
        libraries = [
            LibraryReference.parse_from_table(lib_ref_table) for lib_ref_table in compiled_with_table["libraries"]
        ]
        return CompiledWith(
            compiler=compiled_with_table["compiler"],
            flags=compiled_with_table["flags"],
            crt=compiled_with_table["crt"],
            libraries=libraries
        )


@dataclass
class Declaration:
    TableName = "declaration"
    code: str = ""

    def to_table(self):
        table = tomlkit.table()
        table.add("code", self.code)
        return table

    @staticmethod
    def parse_from_table(declaration_table):
        required_table_entries = ["code"]
        _check_required_table_entries(declaration_table, required_table_entries)
        return Declaration(code=declaration_table["code"])


@dataclass
class HATFile:
    """Encapsulates a HAT file. An instance of this class can be created by calling the
    Deserialize class method e.g.:
        some_hat_file = Deserialize('someFile.hat')
    Similarly, HAT files can be serialized but creating/modifying a HATFile instance
    and then calling Serilize e.g.:
        some_hat_file.name = 'some new name'
        some_hat_file.Serialize(`someFile.hat`)
    """
    name: str = ""
    description: Description = None
    _function_table: FunctionTable = None
    _device_function_table: DeviceFunctionTable = None
    functions: list = field(default_factory=list)
    device_functions: list = field(default_factory=list)
    function_map: Dict[str, Function] = field(default_factory=dict)
    device_function_map: Dict[str, Function] = field(default_factory=dict)
    target: Target = None
    dependencies: Dependencies = None
    compiled_with: CompiledWith = None
    declaration: Declaration = None
    path: Path = None

    HATPrologue = "\n#ifndef __{0}__\n#define __{0}__\n\n#ifdef TOML\n"
    HATEpilogue = "\n#endif // TOML\n\n#endif // __{0}__\n"

    def __post_init__(self):
        self.functions = self._function_table.functions
        self.function_map = self._function_table.function_map
        for func in self.functions:
            func.hat_file = self
            func.link_target = Path(self.path).resolve().parent / self.dependencies.link_target

        if not self._device_function_table:
            self._device_function_table = DeviceFunctionTable({})
        self.device_function_map = self._device_function_table.function_map
        self.device_functions = self._device_function_table.functions

    def Serialize(self, filepath=None):
        """Serilizes the HATFile to disk using the file location specified by `filepath`.
        If `filepath` is not specified then the object's `path` attribute is used."""
        if filepath is None:
            filepath = self.path
        root_table = tomlkit.table()
        root_table.add(Description.TableName, self.description.to_table())
        root_table.add(FunctionTable.TableName, self._function_table.to_table())
        if self.device_function_map:
            root_table.add(DeviceFunctionTable.TableName, self._device_function_table.to_table())
        root_table.add(Target.TableName, self.target.to_table())
        root_table.add(Dependencies.TableName, self.dependencies.to_table())
        root_table.add(CompiledWith.TableName, self.compiled_with.to_table())
        root_table.add(Declaration.TableName, self.declaration.to_table())
        with open(filepath, "w") as out_file:
            # MSVC does not allow "." in macro definitions
            name = self.name.replace(".", "_")
            out_file.write(self.HATPrologue.format(name))
            out_file.write(tomlkit.dumps(root_table))
            out_file.write(self.HATEpilogue.format(name))

    @staticmethod
    def Deserialize(filepath) -> "HATFile":
        """Creates an instance of A HATFile class by deserializing the contents of the file at `filepath`"""
        hat_toml = _read_toml_file(filepath)
        name = os.path.splitext(os.path.basename(filepath))[0]
        required_entries = [
            Description.TableName, FunctionTable.TableName, Target.TableName, Dependencies.TableName,
            CompiledWith.TableName, Declaration.TableName
        ]
        _check_required_table_entries(hat_toml, required_entries)
        device_function_table = None
        if DeviceFunctionTable.TableName in hat_toml:
            device_function_table = DeviceFunctionTable.parse_from_table(hat_toml[DeviceFunctionTable.TableName])
        hat_file = HATFile(
            name=name,
            description=Description.parse_from_table(hat_toml[Description.TableName]),
            _function_table=FunctionTable.parse_from_table(hat_toml[FunctionTable.TableName]),
            _device_function_table=device_function_table,
            target=Target.parse_from_table(hat_toml[Target.TableName]),
            dependencies=Dependencies.parse_from_table(hat_toml[Dependencies.TableName]),
            compiled_with=CompiledWith.parse_from_table(hat_toml[CompiledWith.TableName]),
            declaration=Declaration.parse_from_table(hat_toml[Declaration.TableName]),
            path=Path(filepath).resolve()
        )
        return hat_file



#!/usr/bin/env python3
import accera as acc
import numpy as np
import unittest
from hatlib import load, create_dynamic_package


class HAT_test(unittest.TestCase):
    def test_load(self):

        # Generate a HAT package
        A = acc.Array(role=acc.Array.Role.INPUT, shape=(16, 16))
        B = acc.Array(role=acc.Array.Role.INPUT_OUTPUT, shape=(16, 16))

        nest = acc.Nest(shape=(16, 16))
        i, j = nest.get_indices()

        @nest.iteration_logic
        def _():
            B[i, j] += A[i, j]

        package = acc.Package()
        function = package.add(nest, args=(A, B), base_name="test_function")

        for mode in [acc.Package.Mode.RELEASE, acc.Package.Mode.DEBUG]:
            package_name = f"HAT_test_load_{mode.value}"
            package.build(name=package_name, output_dir="test_acccgen", format=acc.Package.Format.HAT_STATIC, mode=mode)

            create_dynamic_package(f"test_acccgen/{package_name}.hat", f"test_acccgen/{package_name}.dyn.hat")

            _, func_dict = load(f"test_acccgen/{package_name}.dyn.hat")

            for name in func_dict.names:
                print(name)

            # create numpy arguments with the correct shape and dtype
            A = np.random.rand(16, 16).astype(np.float32)
            B = np.random.rand(16, 16).astype(np.float32)
            B_ref = B + A

            # find the function by basename
            test_function = func_dict["test_function"]
            test_function(A, B)

            # check for correctness
            np.testing.assert_allclose(B, B_ref)

            # find the function by actual name
            B_ref = B + A
            test_function1 = func_dict[function.name]
            test_function1(A, B)

            # check for correctness
            np.testing.assert_allclose(B, B_ref)



"""Loads a dynamically-linked HAT package in Python

Call 'load' to load a HAT package in Python. After loading, call the HAT
functions using numpy arrays as arguments. The shape, element type, and order
of each numpy array should exactly match the requirements of the HAT function.

For example:
    import numpy as np
    import hatlib as hat

    # load the package
    package = hat.load("my_package.hat")

    # print the function names
    for name in package.names:
        print(name)

    # create numpy arguments with the correct shape, dtype, and order
    A = np.ones([256,32], dtype=np.float32, order="C")
    B = np.ones([32,256], dtype=np.float32, order="C")
    D = np.ones([256,32], dtype=np.float32, order="C")
    E = np.ones([256,32], dtype=np.float32, order="C")

    # call a package function named 'my_func_698b5e5c'
    package.my_func_698b5e5c(A, B, D, E)
"""
import numpy as np
from typing import Tuple, Union
from functools import reduce

from . import hat_file
from . import hat_package
from .arg_info import ArgInfo


def generate_input_sets_for_func(func: hat_file.Function, input_sets_minimum_size_MB: int = 0, num_additional: int = 0):
    parameters = list(map(ArgInfo, func.arguments))
    shapes_to_sizes = [reduce(lambda x, y: x * y, p.numpy_shape) for p in parameters]
    set_size = reduce(lambda x, y: x + y, map(lambda size, p: size * p.element_num_bytes, shapes_to_sizes, parameters))

    num_input_sets = (input_sets_minimum_size_MB * 1024 * 1024 // set_size) + 1 + num_additional

    input_sets = [[
        np.lib.stride_tricks.as_strided(
            np.random.rand(p.total_element_count).astype(p.numpy_dtype),
            shape=p.numpy_shape,
            strides=p.numpy_strides
        ) for p in parameters
    ] for _ in range(num_input_sets)]

    return input_sets[0] if len(input_sets) == 1 else input_sets


def generate_input_sets_for_hat_file(hat_path):
    t = hat_file.HATFile.Deserialize(hat_path)
    return {func_name: generate_input_sets_for_func(func_desc)
            for func_name, func_desc in t.function_map.items()}


def load(hat_path, try_dynamic_load=True) -> Tuple[hat_package.HATPackage, Union[hat_package.AttributeDict, None]]:
    """
    Returns a HATPackage object loaded from the path provided. If
    `try_dynamic_load` is True, a non-empty dictionary object that can be used
    to invoke the functions in the HATPackage on the current system is the
    second returned object, `None` otherwise.
    """

    pkg = hat_package.HATPackage(hat_file_path=hat_path)

    # TODO: Add heuristics to determine whether loading is possible on this system
    function_dict = None

    if try_dynamic_load:
        try:
            function_dict = hat_package.hat_package_to_func_dict(pkg)
        except:
            # TODO: Figure out how to communicate failure better
            pass

    return pkg, function_dict



#!/usr/bin/env python3

import os
import sys
import unittest
from pathlib import Path
from hatlib import (
    CallingConventionType, CompiledWith, Declaration, Dependencies, Description, Function, FunctionTable, HATFile,
    OperatingSystem, Parameter, ParameterType, Target, UsageType
)


class HATFile_test(unittest.TestCase):

    def test_file_basic_serialize(self):
        # Construct a HAT file from scratch
        # Start with a function definition
        my_function = Function(
            name="my_function",
            description="Some description",
            calling_convention=CallingConventionType.StdCall,
            return_info=Parameter(
                logical_type=ParameterType.RuntimeArray,
                declared_type="float*",
                element_type="float",
                usage=UsageType.Input,
                shape="[16, 16]",
                affine_map=[16, 1],
                size="16 * 16 * sizeof(float)"
            )
        )
        # Create the function table
        functions = FunctionTable({"my_function": my_function})
        # Create the HATFile object
        hat_file1 = HATFile(
            name="test_file",
            description=Description(
                version="0.0.1", author="me", license_url="https://www.apache.org/licenses/LICENSE-2.0.html"
            ),
            _function_table=functions,
            target=Target(
                required=Target.Required(
                    os=OperatingSystem.Windows,
                    cpu=Target.Required.CPU(architecture="Haswell", extensions=["AVX2"]),
                    gpu=None
                ),
                optimized_for=Target.OptimizedFor()
            ),
            dependencies=Dependencies(link_target="my_lib.lib"),
            compiled_with=CompiledWith(compiler="VC++"),
            declaration=Declaration(),
            path=Path(".").resolve()
        )
        # Serialize it to disk
        test_file_name = "test_file_serialize.hat"

        try:
            hat_file1.Serialize(test_file_name)
            # Deserialize it and verify it has what we expect
            hat_file2 = HATFile.Deserialize(test_file_name)
        finally:
            # Remove the file
            os.remove(test_file_name)

        # Do basic verification that the deserialized HatFile contains what we
        # specified when we created the HATFile directly
        self.assertEqual(hat_file1.description, hat_file2.description)
        self.assertEqual(hat_file1.dependencies, hat_file2.dependencies)
        self.assertEqual(hat_file1.compiled_with.to_table(), hat_file2.compiled_with.to_table())
        self.assertTrue("my_function" in hat_file2.function_map)

    def test_file_basic_deserialize(self):
        # Load a HAT file from the samples directory
        hat_file1 = HATFile.Deserialize(
            os.path.join(os.path.dirname(__file__), "..", "samples", "sample_gemm_library.hat")
        )
        description = {
            "author": "John Doe",
            "version": "1.2.3.5",
            "license_url": "https://www.apache.org/licenses/LICENSE-2.0.html",
        }

        # Do basic verification of known values in the file
        # Verify the description has entries we expect
        self.assertLessEqual(description.items(), hat_file1.description.to_table().items())
        # Verify the list of functions
        self.assertTrue(len(hat_file1.functions) == 2)
        self.assertTrue("GEMM_B94D27B9934D3E08" in hat_file1.function_map)
        self.assertTrue("blas_sgemm_row_major" in hat_file1.function_map)


if __name__ == '__main__':
    unittest.main()



from typing import Dict

# lifted from https://github.com/NVIDIA/jitify/blob/master/jitify.hpp
CUDA_HEADER_MAP: Dict[str, str] = {
    'float.h':
        """
#pragma once
#define FLT_RADIX       2
#define FLT_MANT_DIG    24
#define DBL_MANT_DIG    53
#define FLT_DIG         6
#define DBL_DIG         15
#define FLT_MIN_EXP     -125
#define DBL_MIN_EXP     -1021
#define FLT_MIN_10_EXP  -37
#define DBL_MIN_10_EXP  -307
#define FLT_MAX_EXP     128
#define DBL_MAX_EXP     1024
#define FLT_MAX_10_EXP  38
#define DBL_MAX_10_EXP  308
#define FLT_MAX         3.4028234e38f
#define DBL_MAX         1.7976931348623157e308
#define FLT_EPSILON     1.19209289e-7f
#define DBL_EPSILON     2.220440492503130e-16
#define FLT_MIN         1.1754943e-38f
#define DBL_MIN         2.2250738585072013e-308
#define FLT_ROUNDS      1
#if defined __cplusplus && __cplusplus >= 201103L
#define FLT_EVAL_METHOD 0
#define DECIMAL_DIG     21
#endif

""",
    'limits.h':
        """
#pragma once
#if defined _WIN32 || defined _WIN64
 #define __WORDSIZE 32
#else
 #if defined __x86_64__ && !defined __ILP32__
  #define __WORDSIZE 64
 #else
  #define __WORDSIZE 32
 #endif
#endif
#define MB_LEN_MAX  16
#define CHAR_BIT    8
#define SCHAR_MIN   (-128)
#define SCHAR_MAX   127
#define UCHAR_MAX   255
enum {
  _JITIFY_CHAR_IS_UNSIGNED = (char)-1 >= 0,
  CHAR_MIN = _JITIFY_CHAR_IS_UNSIGNED ? 0 : SCHAR_MIN,
  CHAR_MAX = _JITIFY_CHAR_IS_UNSIGNED ? UCHAR_MAX : SCHAR_MAX,
};
#define SHRT_MIN    (-32768)
#define SHRT_MAX    32767
#define USHRT_MAX   65535
#define INT_MIN     (-INT_MAX - 1)
#define INT_MAX     2147483647
#define UINT_MAX    4294967295U
#if __WORDSIZE == 64
 # define LONG_MAX  9223372036854775807L
#else
 # define LONG_MAX  2147483647L
#endif
#define LONG_MIN    (-LONG_MAX - 1L)
#if __WORDSIZE == 64
 #define ULONG_MAX  18446744073709551615UL
#else
 #define ULONG_MAX  4294967295UL
#endif
#define LLONG_MAX  9223372036854775807LL
#define LLONG_MIN  (-LLONG_MAX - 1LL)
#define ULLONG_MAX 18446744073709551615ULL

""",
    'stdint.h':
        """
#pragma once
#include <climits>
namespace __jitify_stdint_ns {
typedef signed char      int8_t;
typedef signed short     int16_t;
typedef signed int       int32_t;
typedef signed long long int64_t;
typedef signed char      int_fast8_t;
typedef signed short     int_fast16_t;
typedef signed int       int_fast32_t;
typedef signed long long int_fast64_t;
typedef signed char      int_least8_t;
typedef signed short     int_least16_t;
typedef signed int       int_least32_t;
typedef signed long long int_least64_t;
typedef signed long long intmax_t;
typedef signed long      intptr_t; //optional
typedef unsigned char      uint8_t;
typedef unsigned short     uint16_t;
typedef unsigned int       uint32_t;
typedef unsigned long long uint64_t;
typedef unsigned char      uint_fast8_t;
typedef unsigned short     uint_fast16_t;
typedef unsigned int       uint_fast32_t;
typedef unsigned long long uint_fast64_t;
typedef unsigned char      uint_least8_t;
typedef unsigned short     uint_least16_t;
typedef unsigned int       uint_least32_t;
typedef unsigned long long uint_least64_t;
typedef unsigned long long uintmax_t;
#define INT8_MIN    SCHAR_MIN
#define INT16_MIN   SHRT_MIN
#if defined _WIN32 || defined _WIN64
#define WCHAR_MIN   0
#define WCHAR_MAX   USHRT_MAX
typedef unsigned long long uintptr_t; //optional
#else
#define WCHAR_MIN   INT_MIN
#define WCHAR_MAX   INT_MAX
typedef unsigned long      uintptr_t; //optional
#endif
#define INT32_MIN   INT_MIN
#define INT64_MIN   LLONG_MIN
#define INT8_MAX    SCHAR_MAX
#define INT16_MAX   SHRT_MAX
#define INT32_MAX   INT_MAX
#define INT64_MAX   LLONG_MAX
#define UINT8_MAX   UCHAR_MAX
#define UINT16_MAX  USHRT_MAX
#define UINT32_MAX  UINT_MAX
#define UINT64_MAX  ULLONG_MAX
#define INTPTR_MIN  LONG_MIN
#define INTMAX_MIN  LLONG_MIN
#define INTPTR_MAX  LONG_MAX
#define INTMAX_MAX  LLONG_MAX
#define UINTPTR_MAX ULONG_MAX
#define UINTMAX_MAX ULLONG_MAX
#define PTRDIFF_MIN INTPTR_MIN
#define PTRDIFF_MAX INTPTR_MAX
#define SIZE_MAX    UINT64_MAX
} // namespace __jitify_stdint_ns
namespace std { using namespace __jitify_stdint_ns; }
using namespace __jitify_stdint_ns;

""",
    'math.h':
        """
#pragma once
namespace __jitify_math_ns {
#if __cplusplus >= 201103L
#define DEFINE_MATH_UNARY_FUNC_WRAPPER(f) \\
    inline double      f(double x)         { return ::f(x); } \\
    inline float       f##f(float x)       { return ::f(x); } \\
    /*inline long double f##l(long double x) { return ::f(x); }*/ \\
    inline float       f(float x)          { return ::f(x); } \\
    /*inline long double f(long double x)    { return ::f(x); }*/
#else
#define DEFINE_MATH_UNARY_FUNC_WRAPPER(f) \\
    inline double      f(double x)         { return ::f(x); } \\
    inline float       f##f(float x)       { return ::f(x); } \\
    /*inline long double f##l(long double x) { return ::f(x); }*/
#endif
DEFINE_MATH_UNARY_FUNC_WRAPPER(cos)
DEFINE_MATH_UNARY_FUNC_WRAPPER(sin)
DEFINE_MATH_UNARY_FUNC_WRAPPER(tan)
DEFINE_MATH_UNARY_FUNC_WRAPPER(acos)
DEFINE_MATH_UNARY_FUNC_WRAPPER(asin)
DEFINE_MATH_UNARY_FUNC_WRAPPER(atan)
template<typename T> inline T atan2(T y, T x) { return ::atan2(y, x); }
DEFINE_MATH_UNARY_FUNC_WRAPPER(cosh)
DEFINE_MATH_UNARY_FUNC_WRAPPER(sinh)
DEFINE_MATH_UNARY_FUNC_WRAPPER(tanh)
DEFINE_MATH_UNARY_FUNC_WRAPPER(exp)
template<typename T> inline T frexp(T x, int* exp) { return ::frexp(x, exp); }
template<typename T> inline T ldexp(T x, int  exp) { return ::ldexp(x, exp); }
DEFINE_MATH_UNARY_FUNC_WRAPPER(log)
DEFINE_MATH_UNARY_FUNC_WRAPPER(log10)
template<typename T> inline T modf(T x, T* intpart) { return ::modf(x, intpart); }
template<typename T> inline T pow(T x, T y) { return ::pow(x, y); }
DEFINE_MATH_UNARY_FUNC_WRAPPER(sqrt)
DEFINE_MATH_UNARY_FUNC_WRAPPER(ceil)
DEFINE_MATH_UNARY_FUNC_WRAPPER(floor)
template<typename T> inline T fmod(T n, T d) { return ::fmod(n, d); }
DEFINE_MATH_UNARY_FUNC_WRAPPER(fabs)
template<typename T> inline T abs(T x) { return ::abs(x); }
#if __cplusplus >= 201103L
DEFINE_MATH_UNARY_FUNC_WRAPPER(acosh)
DEFINE_MATH_UNARY_FUNC_WRAPPER(asinh)
DEFINE_MATH_UNARY_FUNC_WRAPPER(atanh)
DEFINE_MATH_UNARY_FUNC_WRAPPER(exp2)
DEFINE_MATH_UNARY_FUNC_WRAPPER(expm1)
template<typename T> inline int ilogb(T x) { return ::ilogb(x); }
DEFINE_MATH_UNARY_FUNC_WRAPPER(log1p)
DEFINE_MATH_UNARY_FUNC_WRAPPER(log2)
DEFINE_MATH_UNARY_FUNC_WRAPPER(logb)
template<typename T> inline T scalbn (T x, int n)  { return ::scalbn(x, n); }
template<typename T> inline T scalbln(T x, long n) { return ::scalbn(x, n); }
DEFINE_MATH_UNARY_FUNC_WRAPPER(cbrt)
template<typename T> inline T hypot(T x, T y) { return ::hypot(x, y); }
DEFINE_MATH_UNARY_FUNC_WRAPPER(erf)
DEFINE_MATH_UNARY_FUNC_WRAPPER(erfc)
DEFINE_MATH_UNARY_FUNC_WRAPPER(tgamma)
DEFINE_MATH_UNARY_FUNC_WRAPPER(lgamma)
DEFINE_MATH_UNARY_FUNC_WRAPPER(trunc)
DEFINE_MATH_UNARY_FUNC_WRAPPER(round)
template<typename T> inline long lround(T x) { return ::lround(x); }
template<typename T> inline long long llround(T x) { return ::llround(x); }
DEFINE_MATH_UNARY_FUNC_WRAPPER(rint)
template<typename T> inline long lrint(T x) { return ::lrint(x); }
template<typename T> inline long long llrint(T x) { return ::llrint(x); }
DEFINE_MATH_UNARY_FUNC_WRAPPER(nearbyint)
// TODO: remainder, remquo, copysign, nan, nextafter, nexttoward, fdim,
// fmax, fmin, fma
#endif
#undef DEFINE_MATH_UNARY_FUNC_WRAPPER
} // namespace __jitify_math_ns
namespace std { using namespace __jitify_math_ns; }
#define M_PI 3.14159265358979323846
// Note: Global namespace already includes CUDA math funcs
//using namespace __jitify_math_ns;

""",
    'cuda_fp16.h': "",
}

CUDA_HEADER_MAP['climits'] = CUDA_HEADER_MAP['limits.h']

ROCM_HEADER_MAP: Dict[str, str] = {}



#!/usr/bin/env python3

import argparse
import numpy as np
import pandas as pd
import sys
import time
import traceback

from .callable_func import CallableFunc
from .hat_file import HATFile
from .hat import load, generate_input_sets_for_func


class Benchmark:
    """A basic python-based benchmark.
    Useful for comparison only, due to overhead in the Python layer.

    Requirements:
        A compilation toolchain in your PATH: cl.exe & link.exe (Windows), gcc (Linux), or clang (macOS)
    """
    def __init__(self, hat_path: str):
        self.hat_path = hat_path
        self.hat_package, self.func_dict = load(self.hat_path)
        self.hat_functions = self.func_dict.names

        # create dictionary of function descriptions defined in the hat file
        self.function_descriptions = self.hat_package.hat_file.function_map

    def run(self,
            function_name: str,
            warmup_iterations: int = 10,
            min_timing_iterations: int = 100,
            batch_size: int = 10,
            min_time_in_sec: int = 10,
            input_sets_minimum_size_MB=50,
            gpu_id: int=0,
            verbose: bool=False) -> float:
        """Runs benchmarking for a function.
           Multiple inputs are run through the function until both minimum time and minimum iterations have been reached.
           The mean duration is then calculated as mean_duration = total_time_elapsed / total_iterations_performed.
        Args:
            function_name: name of the function
            warmup_iterations: number of warmup iterations
            min_timing_iterations: minimum number of timing iterations
            min_time_in_sec: minimum amount of time to run the benchmark
            input_sets_minimum_size_MB: generate enough input sets to exceed this size to avoid cache hits
            gpu_id: the ID of the gpu device on which to run the benchmark
        Returns:
            Mean duration in seconds,
            Vector of timings in seconds for each batch that was run
        """
        if function_name not in self.hat_functions:
            raise ValueError(f"{function_name} is not found")

        # TODO: support packing and unpacking functions

        mean_elapsed_time, batch_timings = self._profile(
            function_name, warmup_iterations, min_timing_iterations, batch_size,
            min_time_in_sec, input_sets_minimum_size_MB, gpu_id, verbose)

        if verbose:
            print(f"[Benchmarking] Mean duration per iteration: {mean_elapsed_time:.8f}s")

        return mean_elapsed_time, batch_timings

    def _profile(self, function_name, warmup_iterations, min_timing_iterations, batch_size,
                 min_time_in_sec, input_sets_minimum_size_MB, gpu_id: int, verbose: bool):
        def get_perf_counter():
            if hasattr(time, 'perf_counter_ns'):
                _perf_counter = time.perf_counter_ns
                perf_counter_scale = 1000000000
            else:
                _perf_counter = time.perf_counter
                perf_counter_scale = 1
            def perf_counter():
                return _perf_counter() / perf_counter_scale

            return perf_counter

        func = self.function_descriptions[function_name]

        benchmark_func = self.func_dict[function_name]
        if not isinstance(benchmark_func, CallableFunc):
            # generate sufficient input sets to overflow the L3 cache, since we don't know the size of the model
            # we'll make a guess based on the minimum input set size
            input_sets = generate_input_sets_for_func(func,
                                                    input_sets_minimum_size_MB,
                                                    num_additional=10)

            set_size = 0
            for i in input_sets[0]:
                set_size += i.size * i.dtype.itemsize

            if verbose:
                print(f"[Benchmarking] Using {len(input_sets)} input sets, each {set_size} bytes")

            perf_counter = get_perf_counter()
            if verbose:
                print(f"[Benchmarking] Warming up for {warmup_iterations} iterations...")

            for _ in range(warmup_iterations):
                for calling_args in input_sets:
                    self.func_dict[function_name](*calling_args)

            if verbose:
                print(f"[Benchmarking] Timing for at least {min_time_in_sec}s and at least {min_timing_iterations} iterations...")

            start_time_secs = perf_counter()
            end_time_secs = perf_counter()

            i = 0
            i_max = len(input_sets)
            iterations = 1
            batch_timings = []
            while ((end_time_secs - start_time_secs)) < min_time_in_sec and len(batch_timings) < batch_size:
                batch_start_time_secs = perf_counter()
                for _ in range(min_timing_iterations):
                    self.func_dict[function_name](*input_sets[i])
                    i = iterations % i_max
                    iterations += 1
                end_time_secs = perf_counter()
                batch_timings.append(
                    (end_time_secs - batch_start_time_secs))

            elapsed_time_secs = ((end_time_secs - start_time_secs))
            mean_elapsed_time_secs = elapsed_time_secs / iterations
            return mean_elapsed_time_secs, batch_timings
        else:
            if verbose:
                print(f"[Benchmarking] Benchmarking device function on gpu {gpu_id}. {batch_size} batches of warming up for {warmup_iterations} and then measuring with {min_timing_iterations} iterations.")
            input_sets = generate_input_sets_for_func(func)

            set_size = 0
            for i in input_sets:
                set_size += i.size * i.dtype.itemsize

            if verbose:
                print(f"[Benchmarking] Using input of {set_size} bytes")

            batch_timings_ms = benchmark_func.benchmark(warmup_iters=warmup_iterations, iters=min_timing_iterations, batch_size=batch_size, args=input_sets, gpu_id=gpu_id)
            batch_timings_secs = list(map(lambda t: t / 1000, batch_timings_ms))
            mean_timings = sum(batch_timings_secs) / (min_timing_iterations * batch_size)
            return mean_timings, batch_timings_secs


def write_runtime_to_hat_file(hat_path, function_name, mean_time_secs):
    """Writes the mean time in seconds to a HAT file
    """
    # Write back the runtime to the HAT file
    hat_file = HATFile.Deserialize(hat_path)
    hat_func = hat_file.function_map.get(function_name)
    hat_func.auxiliary["mean_duration_in_sec"] = mean_time_secs

    hat_file.Serialize(hat_path)

    # Workaround to remove extra empty lines
    with open(hat_path, "r") as f:
        lines = f.readlines()
        lines = [
            lines[i] for i in range(len(lines))
            if not (lines[i] == "\n" and i < len(lines) -
                    1 and lines[i + 1] == "\n")
        ]
    with open(hat_path, "w") as f:
        f.writelines(lines)


def run_benchmark(hat_path,
                  store_in_hat=False,
                  batch_size=10,
                  min_time_in_sec=10,
                  input_sets_minimum_size_MB=50,
                  gpu_id: int=0,
                  verbose: bool=False):
    results = []

    benchmark = Benchmark(hat_path)
    functions = benchmark.hat_functions
    for function_name in functions:
        if verbose:
            print(f"\nBenchmarking function: {function_name}")
        if "Initialize" in function_name or "_debug_check_allclose" in function_name:  # Skip init and debug functions
            continue

        try:
            _, batch_timings = benchmark.run(
                function_name,
                warmup_iterations=batch_size,
                min_timing_iterations=batch_size,
                batch_size=batch_size,
                min_time_in_sec=min_time_in_sec,
                input_sets_minimum_size_MB=input_sets_minimum_size_MB,
                gpu_id=gpu_id,
                verbose=verbose)

            sorted_batch_means = np.array(sorted(batch_timings)) / batch_size
            num_batches = len(batch_timings)

            mean_of_means = sorted_batch_means.mean()
            median_of_means = sorted_batch_means[num_batches // 2]
            mean_of_small_means = sorted_batch_means[0:num_batches // 2].mean()
            robust_means = sorted_batch_means[(num_batches //
                                               5):(-num_batches // 5)]
            robust_mean_of_means = robust_means.mean()
            min_of_means = sorted_batch_means[0]

            if store_in_hat:
                write_runtime_to_hat_file(hat_path, function_name,
                                          mean_of_means)
            results.append({
                "function_name": function_name,
                "mean": mean_of_means,
                "median_of_means": median_of_means,
                "mean_of_small_means": mean_of_small_means,
                "robust_mean": robust_mean_of_means,
                "min_of_means": min_of_means,
            })
        except Exception as e:
            if verbose:
                exc_type, exc_val, exc_tb = sys.exc_info()
                traceback.print_exception(exc_type,
                                        exc_val,
                                        exc_tb,
                                        file=sys.stderr)
                print("\nException message: ", e)
                print(f"WARNING: Failed to run function {function_name}, skipping this benchmark.")

            raise e
    return results


def main(argv):
    arg_parser = argparse.ArgumentParser(
        description=
        "Benchmarks each function in a HAT package and estimates its duration.\n"
        "Example:\n"
        "    hatlib.benchmark_hat_package <hat_path>\n")

    arg_parser.add_argument("hat_path",
                            help="Path to the HAT file",
                            default=None)
    arg_parser.add_argument(
        "--store_in_hat",
        help=
        "If set, will write the duration as meta-data back into the hat file",
        action='store_true')
    arg_parser.add_argument("--results_file",
                            help="Full path where the results will be written",
                            default="results.csv")
    arg_parser.add_argument(
        "--batch_size",
        help=
        "The number of function calls in each batch (at least one full batch is executed)",
        default=10)
    arg_parser.add_argument(
        "--min_time_in_sec",
        help="Minimum number of seconds to run the benchmark for",
        default=30)
    arg_parser.add_argument(
        "--input_sets_minimum_size_MB",
        help=
        "Minimum size in MB of the input sets. Typically this is large enough to ensure eviction of the biggest cache on the target (e.g. L3 on an desktop CPU)",
        default=50)
    arg_parser.add_argument(
        "--verbose",
        help="Enable verbose logging",
        default=False)

    args = vars(arg_parser.parse_args(argv))

    results = run_benchmark(args["hat_path"],
                            args["store_in_hat"],
                            batch_size=int(args["batch_size"]),
                            min_time_in_sec=int(args["min_time_in_sec"]),
                            input_sets_minimum_size_MB=int(args["input_sets_minimum_size_MB"]),
                            verbose=bool(args["verbose"]))
    df = pd.DataFrame(results)
    df.to_csv(args["results_file"], index=False)
    pd.options.display.float_format = '{:8.8f}'.format
    print(df)

    print(f"Results saved to {args['results_file']}")


def main_command():
    main(sys.argv[1:])  # drop the first argument (program name)


if __name__ == "__main__":
    main_command()



"""
Python interface to hiprtc
lifted from https://github.com/jatinx/PyHIP
TODO: move to a submodule
"""

import ctypes
import sys

# _libhiprtc_libname = 'libhiprtc.so' # Currently its the same library
_libhiprtc_libname = 'libamdhip64.so'

_libhiprtc = None
if 'linux' in sys.platform:
    _libhiprtc = ctypes.cdll.LoadLibrary(_libhiprtc_libname)
else:
    # Currently we do not support windows, mainly because I do not have a windows build of hip
    raise RuntimeError('Only linux is supported')

if _libhiprtc is None:
    raise OSError('hiprtc library not found')


def POINTER(obj):
    """
    ctype pointer to object
    """
    p = ctypes.POINTER(obj)
    if not isinstance(p.from_param, classmethod):

        def from_param(cls, x):
            if x is None:
                return cls()
            else:
                return x

        p.from_param = classmethod(from_param)

    return p


_libhiprtc.hiprtcGetErrorString.restype = ctypes.c_char_p
_libhiprtc.hiprtcGetErrorString.argtypes = [ctypes.c_int]


def hiprtcGetErrorString(e):
    """
    Retrieve hiprtc error string.

    Return the string associated with the specified hiprtc error status
    code.

    Parameters
    ----------
    e : int
        Error number.

    Returns
    -------
    s : str
        Error string.

    """

    return _libhiprtc.hiprtcGetErrorString(e)


# Generic hiprtc Error
class hiprtcError(Exception):
    """hiprtc error."""
    pass


class hiprtcErrorOutOfMemory(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(1)
    pass


class hiprtcErrorProgramCreationFailure(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(2)
    pass


class hiprtcErrorInvalidInput(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(3)
    pass


class hiprtcErrorInvalidProgram(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(4)
    pass


class hiprtcErrorInvalidOption(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(5)
    pass


class hiprtcErrorCompilation(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(6)
    pass


class hiprtcErrorBuiltinOperationFailure(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(7)
    pass


class hiprtcErrorNoNameExpressionAfterCompilation(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(8)
    pass


class hiprtcErrorNoLoweredNamesBeforeCompilation(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(9)
    pass


class hiprtcErrorNameExpressionNotValid(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(10)
    pass


class hiprtcErrorInternalError(hiprtcError):
    __doc__ = _libhiprtc.hiprtcGetErrorString(11)
    pass


hiprtcExceptions = {
    1: hiprtcErrorOutOfMemory,
    2: hiprtcErrorProgramCreationFailure,
    3: hiprtcErrorInvalidInput,
    4: hiprtcErrorInvalidProgram,
    5: hiprtcErrorInvalidOption,
    6: hiprtcErrorCompilation,
    7: hiprtcErrorBuiltinOperationFailure,
    8: hiprtcErrorNoNameExpressionAfterCompilation,
    9: hiprtcErrorNoLoweredNamesBeforeCompilation,
    10: hiprtcErrorNameExpressionNotValid,
    11: hiprtcErrorInternalError
}


def hiprtcCheckStatus(status):
    if status != 0:
        try:
            e = hiprtcExceptions[status]
        except KeyError:
            raise hiprtcError('unknown hiprtc error %s' % status)
        else:
            raise e


_libhiprtc.hiprtcCreateProgram.restype = int
_libhiprtc.hiprtcCreateProgram.argtypes = [
    ctypes.POINTER(ctypes.c_void_p),  # hiprtcProgram
    ctypes.POINTER(ctypes.c_char),  # Source
    ctypes.POINTER(ctypes.c_char),  # Name
    ctypes.c_int,  # numberOfHeaders
    ctypes.POINTER(ctypes.c_char_p),  # header
    ctypes.POINTER(ctypes.c_char_p)
]  # headerNames


def hiprtcCreateProgram(source, name, header_names, header_sources):
    """
    Create hiprtcProgram

    Parameters
    ----------
    source : string
        Source in python string
    name : string
        Program name
    header_names: list of string
        list of headernames
    header_sources: list of string
        list of headernames

    Returns
    -------
    prog : ctypes pointer
        hiprtc program handle
    """

    # Encode strings to utf-8
    e_source = source.encode('utf-8')
    e_name = name.encode('utf-8')
    e_header_names = list()
    e_header_sources = list()
    for header_name in header_names:
        e_header_name = header_name.encode('utf-8')
        e_header_names.append(e_header_name)
    for header_source in header_sources:
        e_header_source = header_source.encode('utf-8')
        e_header_sources.append(e_header_source)

    prog = ctypes.c_void_p()
    c_header_names = (ctypes.c_char_p * len(e_header_names))()
    c_header_names[:] = e_header_names
    c_header_sources = (ctypes.c_char_p * len(e_header_sources))()
    c_header_sources[:] = e_header_sources

    status = _libhiprtc.hiprtcCreateProgram(ctypes.byref(prog), e_source,
                                            e_name, len(e_header_names),
                                            c_header_sources, c_header_names)
    hiprtcCheckStatus(status)
    return prog


_libhiprtc.hiprtcDestroyProgram.restype = int
_libhiprtc.hiprtcDestroyProgram.argtypes = [ctypes.POINTER(ctypes.c_void_p)
                                            ]  # hiprtcProgram


def hiprtcDestroyProgram(prog):
    """
    Destroy hiprtcProgram

    Parameters
    ----------
    prog : ctypes pointer
        hiprtc program handle
    """
    status = _libhiprtc.hiprtcDestroyProgram(ctypes.byref(prog))
    hiprtcCheckStatus(status)


_libhiprtc.hiprtcAddNameExpression.restype = int
_libhiprtc.hiprtcAddNameExpression.argtypes = [
    ctypes.c_void_p,  # hiprtcProgram
    ctypes.POINTER(ctypes.c_char)
]  # expression


def hiprtcAddNameExpression(prog, expression):
    """
    Tracks expression name through hiprtc compilation unit

    Parameters
    ----------
    prog : ctypes pointer
        hiprtc program handle
    expression : string
        expression name
    """
    e_expression = expression.encode('utf-8')
    status = _libhiprtc.hiprtcAddNameExpression(prog, e_expression)
    hiprtcCheckStatus(status)


_libhiprtc.hiprtcCompileProgram.restype = int
_libhiprtc.hiprtcCompileProgram.argtypes = [
    ctypes.c_void_p,  # hiprtcProgram
    ctypes.c_int,  # num of options
    ctypes.POINTER(ctypes.c_char_p)
]  # options


def hiprtcCompileProgram(prog, options):
    """
    Compiles the hiprtc program

    Parameters
    ----------
    prog : ctypes pointer
        hiprtc program handle
    options : list of string
        option list to be passed to compilation
    """

    e_options = list()
    for option in options:
        e_options.append(option.encode('utf-8'))
    c_options = (ctypes.c_char_p * len(e_options))()
    c_options[:] = e_options
    status = _libhiprtc.hiprtcCompileProgram(prog, len(c_options), c_options)
    if status != 0:
        print(hiprtcGetProgramLog(prog))
    hiprtcCheckStatus(status)


_libhiprtc.hiprtcGetProgramLogSize.restype = int
_libhiprtc.hiprtcGetProgramLogSize.argtypes = [
    ctypes.c_void_p,  # hiprtcProgram
    ctypes.POINTER(ctypes.c_size_t)
]  # Size of log
_libhiprtc.hiprtcGetProgramLog.restype = int
_libhiprtc.hiprtcGetProgramLog.argtypes = [
    ctypes.c_void_p,  # hiprtcProgram
    ctypes.POINTER(ctypes.c_char)
]  # log


def hiprtcGetProgramLog(prog):
    """
    Gets the hiprtc Log

    Parameters
    ----------
    prog : ctypes pointer
        hiprtc program handle

    Returns
    -------
    log : string
        program compilation log (warnings/errors)
    """
    log_size = ctypes.c_size_t()
    status = _libhiprtc.hiprtcGetProgramLogSize(prog, ctypes.byref(log_size))
    hiprtcCheckStatus(status)
    log = "0" * log_size.value
    e_log = log.encode('utf-8')
    status = _libhiprtc.hiprtcGetProgramLog(prog, e_log)
    hiprtcCheckStatus(status)
    return e_log.decode('utf-8')


_libhiprtc.hiprtcGetCodeSize.restype = int
_libhiprtc.hiprtcGetCodeSize.argtypes = [
    ctypes.c_void_p,  # hiprtcProgram
    ctypes.POINTER(ctypes.c_size_t)
]  # Size of log
_libhiprtc.hiprtcGetCode.restype = int
_libhiprtc.hiprtcGetCode.argtypes = [
    ctypes.c_void_p,  # hiprtcProgram
    ctypes.POINTER(ctypes.c_char)
]  # log


def hiprtcGetCode(prog):
    """
    Gets the hiprtc compiled code

    Parameters
    ----------
    prog : ctypes pointer
        hiprtc program handle

    Returns
    -------
    code : string
        hiprtc module code
    """
    code_size = ctypes.c_size_t()
    status = _libhiprtc.hiprtcGetCodeSize(prog, ctypes.byref(code_size))
    hiprtcCheckStatus(status)

    code = "0" * code_size.value
    e_code = code.encode('utf-8')
    status = _libhiprtc.hiprtcGetCode(prog, e_code)
    hiprtcCheckStatus(status)
    return e_code



import os
import pathlib
import sys
import numpy as np
from typing import List
from cuda import cuda, nvrtc
from .arg_info import ArgInfo, verify_args
from .callable_func import CallableFunc
from .hat_file import Function


def ASSERT_DRV(err):
    if isinstance(err, cuda.CUresult):
        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError("Cuda Error: {}".format(cuda.cuGetErrorString(err)[1].decode('utf-8')))
    elif isinstance(err, nvrtc.nvrtcResult):
        if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
            raise RuntimeError("Nvrtc Error: {}".format(err))
    else:
        raise RuntimeError("Unknown error type: {}".format(err))


def _find_cuda_incl_path() -> pathlib.Path:
    "Tries to find the CUDA include path."
    cuda_path = os.getenv("CUDA_PATH")
    if cuda_path is None:
        if sys.platform == 'linux':
            cuda_path = pathlib.Path("/usr/local/cuda/include")
            if not (cuda_path.exists() and cuda_path.is_dir()):
                cuda_path = None
        elif sys.platform == 'win32':
            ...
        elif sys.platform == 'darwin':
            ...
    else:
        cuda_path = pathlib.Path(cuda_path)
        cuda_path /= "include"

    return cuda_path

def _get_compute_capability(gpu_id) -> int:
    err, major = cuda.cuDeviceGetAttribute(cuda.CUdevice_attribute.CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MAJOR, gpu_id)
    ASSERT_DRV(err)

    err, minor = cuda.cuDeviceGetAttribute(cuda.CUdevice_attribute.CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MINOR, gpu_id)
    ASSERT_DRV(err)

    return (major * 10) + minor

def compile_cuda_program(cuda_src_path: pathlib.Path, func_name, gpu_id):
    src = cuda_src_path.read_text()

    cuda_incl_path = _find_cuda_incl_path()
    if not cuda_incl_path:
        raise RuntimeError("Unable to determine CUDA include path. Please set CUDA_PATH environment variable.")

    opts = [
        # https://docs.nvidia.com/cuda/nvrtc/index.html#group__options
        f'--gpu-architecture=compute_{_get_compute_capability(gpu_id)}'.encode(),
        b'--ptxas-options=--warn-on-spills', # https://docs.nvidia.com/cuda/cuda-compiler-driver-nvcc/index.html#options-for-passing-specific-phase-options-ptxas-options
        b'-use_fast_math',
        b'--include-path=' + str(cuda_incl_path).encode(),
        b'-std=c++17',
        b'-default-device',
        #b'--restrict',
        #b'--device-int128'
    ]

    # Create program
    err, prog = nvrtc.nvrtcCreateProgram(str.encode(src), func_name.encode('utf-8'), 0, [], [])
    ASSERT_DRV(err)

    # Compile program
    err = nvrtc.nvrtcCompileProgram(prog, len(opts), opts)
    if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
        err, log_size = nvrtc.nvrtcGetProgramLogSize(prog)
        ASSERT_DRV(err)

        log = "0" * log_size
        e_log = log.encode('utf-8')
        err = nvrtc.nvrtcGetProgramLog(prog, e_log)
        print(e_log.decode('utf-8'))

    # Get PTX from compilation
    err, ptxSize = nvrtc.nvrtcGetPTXSize(prog)
    ASSERT_DRV(err)
    ptx = b" " * ptxSize
    err = nvrtc.nvrtcGetPTX(prog, ptx)

    return ptx


def initialize_cuda(gpu_id):
    # Initialize CUDA Driver API
    err, = cuda.cuInit(0)
    ASSERT_DRV(err)

    err, cuDevice = cuda.cuDeviceGet(gpu_id)
    ASSERT_DRV(err)

    # Create context
    # TODO: USE cuDevicePrimaryCtxRetain for faster intialization
    err, context = cuda.cuCtxCreate(0, cuDevice)
    ASSERT_DRV(err)
    return context


def get_func_from_ptx(ptx, func_name):
    # Load PTX as module data and retrieve function
    ptx = np.char.array(ptx)
    err, ptx_mod = cuda.cuModuleLoadData(ptx)
    ASSERT_DRV(err)
    err, kernel = cuda.cuModuleGetFunction(ptx_mod, func_name.encode('utf-8'))
    ASSERT_DRV(err)

    return kernel


def _cuda_transfer_mem(usage, func, source_args: List, dest_args: List, arg_infos: List[ArgInfo]):
    for source_arg, dest_arg, arg_info in zip(source_args, dest_args, arg_infos):
        if usage in arg_info.usage.value:
            err, = func(dest_arg, source_arg, arg_info.total_byte_size)
            ASSERT_DRV(err)


def transfer_mem_host_to_cuda(device_args: List, host_args: List[np.array], arg_infos: List[ArgInfo]):
    _cuda_transfer_mem(
        usage='input',
        func=cuda.cuMemcpyHtoD,
        source_args=[a.ctypes.data for a in host_args],
        dest_args=device_args,
        arg_infos=arg_infos
    )


def transfer_mem_cuda_to_host(device_args: List, host_args: List[np.array], arg_infos: List[ArgInfo]):
    _cuda_transfer_mem(
        usage='output',
        func=cuda.cuMemcpyDtoH,
        source_args=device_args,
        dest_args=[a.ctypes.data for a in host_args],
        arg_infos=arg_infos
    )


def allocate_cuda_mem(arg_infos: List[ArgInfo]):
    device_mem = []

    for arg in arg_infos:
        size = arg.total_byte_size
        err, mem = cuda.cuMemAlloc(size)
        try:
            ASSERT_DRV(err)
        except:
            free_cuda_mem(device_mem)
            raise
        device_mem.append(mem)

    return device_mem


def free_cuda_mem(args):
    for arg in args:
        cuda.cuMemFree(arg)


def device_args_to_ptr_list(device_args: List):
    # CUDA python example says this is subject to change
    ptrs = [np.array([int(d_arg)], dtype=np.uint64) for d_arg in device_args]
    ptrs = np.array([ptr.ctypes.data for ptr in ptrs], dtype=np.uint64)

    return ptrs


_PTX_CACHE = {}


class CudaCallableFunc(CallableFunc):

    def __init__(self, func: Function, cuda_src_path: str) -> None:
        super().__init__()
        self.hat_func = func
        self.func_name = func.name
        self.kernel = None
        hat_arg_descriptions = func.arguments
        self.arg_infos = [ArgInfo(d) for d in hat_arg_descriptions]
        self.launch_params = func.launch_parameters
        self.device_mem = None
        self.ptrs = None
        self.start_event = None
        self.stop_event = None
        self.exec_time = 0.
        self.cuda_src_path = cuda_src_path
        self.context = None

    def init_runtime(self, benchmark: bool, gpu_id: int):
        self.context = initialize_cuda(gpu_id)

        ptx = _PTX_CACHE.get(self.cuda_src_path)
        if not ptx:
            _PTX_CACHE[self.cuda_src_path] = ptx = compile_cuda_program(self.cuda_src_path, self.func_name, gpu_id)

        self.kernel = get_func_from_ptx(ptx, self.func_name)

    def cleanup_runtime(self, benchmark: bool):
        cuda.cuCtxDestroy(self.context)

    def init_main(self, benchmark: bool, warmup_iters=0, args=[], gpu_id: int=0):
        verify_args(args, self.arg_infos, self.func_name)
        self.device_mem = allocate_cuda_mem(self.arg_infos)

        if not benchmark:
            transfer_mem_host_to_cuda(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)

        self.ptrs = device_args_to_ptr_list(self.device_mem)

        err, self.start_event = cuda.cuEventCreate(cuda.CUevent_flags.CU_EVENT_DEFAULT)
        ASSERT_DRV(err)
        err, self.stop_event = cuda.cuEventCreate(cuda.CUevent_flags.CU_EVENT_DEFAULT)
        ASSERT_DRV(err)

        for _ in range(warmup_iters):
            err, = cuda.cuLaunchKernel(
                self.kernel,
                *self.launch_params,    # [ grid[x-z], block[x-z] ]
                0,    # dynamic shared memory
                0,    # stream
                self.ptrs.ctypes.data,    # kernel arguments
                0,    # extra (ignore)
            )

            if not benchmark:
                ASSERT_DRV(err)

    def main(self, benchmark: bool, iters=1, batch_size=1, args=[]) -> float:
        batch_timings: List[float] = []
        for _ in range(batch_size):
            err, = cuda.cuEventRecord(self.start_event, 0)
            ASSERT_DRV(err)

            for _ in range(iters):
                err, = cuda.cuLaunchKernel(
                    self.kernel,
                    *self.launch_params,    # [ grid[x-z], block[x-z] ]
                    0,    # dynamic shared memory
                    0,    # stream
                    self.ptrs.ctypes.data,    # kernel arguments
                    0,    # extra (ignore)
                )

                if not benchmark:
                    ASSERT_DRV(err)

            err, = cuda.cuEventRecord(self.stop_event, 0)
            ASSERT_DRV(err)
            err, = cuda.cuEventSynchronize(self.stop_event)
            ASSERT_DRV(err)
            err, batch_time = cuda.cuEventElapsedTime(self.start_event, self.stop_event)
            ASSERT_DRV(err)
            batch_timings.append(batch_time)
            self.exec_time += batch_time

            if not benchmark:
                err, = cuda.cuCtxSynchronize()
                ASSERT_DRV(err)

        self.exec_time /= (iters * batch_size)
        return batch_timings

    def cleanup_main(self, benchmark: bool, args=[]):
        # If there's no device mem, that means allocation during initialization failed, which means nothing else needs to be cleaned up either
        if not benchmark and self.device_mem:
            transfer_mem_cuda_to_host(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)
        if self.device_mem:
            free_cuda_mem(self.device_mem)
        err, = cuda.cuCtxSynchronize()
        ASSERT_DRV(err)

        if self.start_event:
            cuda.cuEventDestroy(self.start_event)

        if self.stop_event:
            cuda.cuEventDestroy(self.stop_event)


def create_loader_for_device_function(device_func: Function, hat_dir_path: str) -> CallableFunc:
    if not device_func.provider:
        raise RuntimeError("Expected a provider for the device function")

    cuda_src_path: pathlib.Path = pathlib.Path(hat_dir_path) / device_func.provider

    return CudaCallableFunc(device_func, cuda_src_path)



#!/usr/bin/env python3
import unittest
import accera as acc
import numpy as np
from hatlib import run_benchmark


class BenchmarkHATPackage_test(unittest.TestCase):
    def test_benchmark(self):
        A = acc.Array(role=acc.Array.Role.INPUT, shape=(256, 256))
        B = acc.Array(role=acc.Array.Role.INPUT, shape=(256, 256))
        C = acc.Array(role=acc.Array.Role.INPUT_OUTPUT, shape=(256, 256))

        nest = acc.Nest(shape=(256, 256, 256))
        i, j, k = nest.get_indices()

        @nest.iteration_logic
        def _():
            C[i, j] += A[i, k] * B[k, j]

        package = acc.Package()
        package.add(nest, args=(A, B, C), base_name="test_function")
        package.build(
            name="BenchmarkHATPackage_test_benchmark", output_dir="test_acccgen", format=acc.Package.Format.HAT_DYNAMIC
        )

        results = run_benchmark(
            "test_acccgen/BenchmarkHATPackage_test_benchmark.hat",
            store_in_hat=False,
            batch_size=2,
            min_time_in_sec=1,
            input_sets_minimum_size_MB=1
        )
        self.assertIn("test_function", results[0]["function_name"])
        self.assertEqual(type(results[0]["mean"]), np.float64)

    def test_benchmark_multiple_functions(self):
        A = acc.Array(role=acc.Array.Role.INPUT, shape=(256, 256))
        B = acc.Array(role=acc.Array.Role.INPUT, shape=(256, 256))
        C = acc.Array(role=acc.Array.Role.INPUT_OUTPUT, shape=(256, 256))
        D = acc.Array(role=acc.Array.Role.INPUT, shape=(256, 256))    # dummy argument

        nest = acc.Nest(shape=(256, 256, 256))
        i, j, k = nest.get_indices()

        @nest.iteration_logic
        def _():
            C[i, j] += A[i, k] * B[k, j]

        package = acc.Package()
        package.add(nest, args=(A, B, C), base_name="test_function")

        # add another function with a dummy argument - run_benchmark should be able to call it
        # with the correct signature
        package.add(nest, args=(A, B, C, D), base_name="test_function_dummy")

        package.build(
            name="BenchmarkHATPackage_test_benchmark", output_dir="test_acccgen", format=acc.Package.Format.HAT_DYNAMIC
        )

        results = run_benchmark(
            "test_acccgen/BenchmarkHATPackage_test_benchmark.hat",
            store_in_hat=False,
            batch_size=2,
            min_time_in_sec=1,
            input_sets_minimum_size_MB=1
        )
        self.assertIn("test_function", results[0]["function_name"])
        self.assertIn("test_function_dummy", results[1]["function_name"])
        for r in results:
            self.assertEqual(type(r["mean"]), np.float64)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import Any, List


class CallableFunc(ABC):

    def __call__(self, *args: Any, gpu_id: int = 0) -> float:
        try:
            self.init_runtime(benchmark=False, gpu_id=gpu_id)
            try:
                self.init_main(benchmark=False, args=args, gpu_id=gpu_id)
                timings: List[float] = self.main(benchmark=False, args=args)
            finally:
                self.cleanup_main(benchmark=False, args=args)
        finally:
            self.cleanup_runtime(benchmark=False)

        return timings[0]

    def benchmark(self, warmup_iters, iters, batch_size, args, gpu_id: int) -> List[float]:
        try:
            self.init_runtime(benchmark=True, gpu_id=gpu_id)
            try:
                self.init_main(benchmark=True, warmup_iters=warmup_iters, args=args, gpu_id=gpu_id)
                timings = self.main(benchmark=True, iters=iters, batch_size=batch_size, args=args)
            finally:
                self.cleanup_main(benchmark=True, args=args)
        finally:
            self.cleanup_runtime(benchmark=True)
        return timings

    @abstractmethod
    def init_runtime(self, benchmark: bool, gpu_id: int):
        ...

    @abstractmethod
    def init_main(self, benchmark: bool, warmup_iters=0, *args) -> float:
        ...

    @abstractmethod
    def main(self, benchmark: bool, iters=1, *args: Any):
        ...

    @abstractmethod
    def cleanup_main(self, benchmark: bool, *args):
        ...

    @abstractmethod
    def cleanup_runtime(self, benchmark: bool):
        ...



#!/usr/bin/env python3

import argparse

from . import hat


def verify_hat_package(hat_path):
    _, funcs = hat.load(hat_path)
    inputs = hat.generate_input_sets_for_hat_file(hat_path)
    for name, fn in funcs.items():
        print(f"\n{'*' * 10}\n")
        
        print(f"[*] Verifying function {name} --")
        func_inputs = inputs[name]

        print("[*] Inputs before function call:")
        for i, func_input in enumerate(func_inputs):
            print(f"[*]\tInput {i}: {','.join(map(str, func_input.ravel()[:32]))}")

        try:
        
            time = fn(*inputs[name])

        except RuntimeError as e:
            print(f"[!] Error while running {name}: {e}")
            continue
        
        print("Inputs after function call:")
        for i, func_input in enumerate(func_inputs):
            print(f"[*]\tInput {i}: {','.join(map(str, func_input.ravel()[:32]))}")

        if time:
            print(f"[*] Function execution time: {time:4f}ms")
        
        del inputs[name]

    else:
        print(f"\n{'*' * 10}\n")


def main():
    arg_parser = argparse.ArgumentParser(
        description=(
            "Executes every available function in the hat package with randomized inputs. Meant for quick verification.\n"
            "Example:\n"
            "    hatlib.verify_hat_package <hat_path>\n"
        )
    )

    arg_parser.add_argument("hat_path", help="Path to the HAT file", default=None)

    args = vars(arg_parser.parse_args())
    verify_hat_package(args["hat_path"])


if __name__ == "__main__":
    main()



from .hat import *
from .hat_file import *
from .hat_package import *
from .hat_to_dynamic import *
from .hat_to_lib import *
from .benchmark_hat_package import run_benchmark
from .platform_utilities import *



"""
Python interface to hip library
lifted from https://github.com/jatinx/PyHIP
TODO: move to a submodule
"""

import ctypes
import sys

_libhip_libname = 'libamdhip64.so'

_libhip = None
if 'linux' in sys.platform:
    _libhip = ctypes.cdll.LoadLibrary(_libhip_libname)
else:
    # Currently we do not support windows, mainly because I do not have a windows build of hip
    raise RuntimeError('Only linux is supported')

if _libhip is None:
    raise OSError('hiprtc library not found')


def POINTER(obj):
    """
    ctype pointer to object
    """
    p = ctypes.POINTER(obj)
    if not isinstance(p.from_param, classmethod):

        def from_param(cls, x):
            if x is None:
                return cls()
            else:
                return x

        p.from_param = classmethod(from_param)

    return p


_libhip.hipGetErrorString.restype = ctypes.c_char_p
_libhip.hipGetErrorString.argtypes = [ctypes.c_int]


def hiprtcGetErrorString(e):
    """
    Retrieve hip error string.

    Return the string associated with the specified hiprtc error status
    code.

    Parameters
    ----------
    e : int
        Error number.

    Returns
    -------
    s : str
        Error string.

    """

    return _libhip.hipGetErrorString(e)


# Generic hip error
class hipError(Exception):
    """hip error"""
    pass


class hipErrorInvalidValue(hipError):
    __doc__ = _libhip.hipGetErrorString(1)
    pass


class hipErrorOutOfMemory(hipError):
    __doc__ = _libhip.hipGetErrorString(2)
    pass


class hipErrorNotInitialized(hipError):
    __doc__ = _libhip.hipGetErrorString(3)
    pass


class hipErrorDeinitialized(hipError):
    __doc__ = _libhip.hipGetErrorString(4)
    pass


class hipErrorProfilerDisabled(hipError):
    __doc__ = _libhip.hipGetErrorString(5)
    pass


class hipErrorProfilerNotInitialized(hipError):
    __doc__ = _libhip.hipGetErrorString(6)
    pass


class hipErrorProfilerAlreadyStarted(hipError):
    __doc__ = _libhip.hipGetErrorString(7)
    pass


class hipErrorProfilerAlreadyStopped(hipError):
    __doc__ = _libhip.hipGetErrorString(8)
    pass


class hipErrorInvalidConfiguration(hipError):
    __doc__ = _libhip.hipGetErrorString(9)
    pass


class hipErrorInvalidSymbol(hipError):
    __doc__ = _libhip.hipGetErrorString(13)
    pass


class hipErrorInvalidDevicePointer(hipError):
    __doc__ = _libhip.hipGetErrorString(17)
    pass


class hipErrorInvalidMemcpyDirection(hipError):
    __doc__ = _libhip.hipGetErrorString(21)
    pass


class hipErrorInsufficientDriver(hipError):
    __doc__ = _libhip.hipGetErrorString(35)
    pass


class hipErrorMissingConfiguration(hipError):
    __doc__ = _libhip.hipGetErrorString(52)
    pass


class hipErrorPriorLaunchFailure(hipError):
    __doc__ = _libhip.hipGetErrorString(53)
    pass


class hipErrorInvalidDeviceFunction(hipError):
    __doc__ = _libhip.hipGetErrorString(98)
    pass


class hipErrorNoDevice(hipError):
    __doc__ = _libhip.hipGetErrorString(100)
    pass


class hipErrorInvalidDevice(hipError):
    __doc__ = _libhip.hipGetErrorString(101)
    pass


class hipErrorInvalidImage(hipError):
    __doc__ = _libhip.hipGetErrorString(200)
    pass


class hipErrorInvalidContext(hipError):
    __doc__ = _libhip.hipGetErrorString(201)
    pass


class hipErrorContextAlreadyCurrent(hipError):
    __doc__ = _libhip.hipGetErrorString(202)
    pass


class hipErrorMapFailed(hipError):
    __doc__ = _libhip.hipGetErrorString(205)
    pass


class hipErrorUnmapFailed(hipError):
    __doc__ = _libhip.hipGetErrorString(206)
    pass


class hipErrorArrayIsMapped(hipError):
    __doc__ = _libhip.hipGetErrorString(207)
    pass


class hipErrorAlreadyMapped(hipError):
    __doc__ = _libhip.hipGetErrorString(208)
    pass


class hipErrorNoBinaryForGpu(hipError):
    __doc__ = _libhip.hipGetErrorString(209)
    pass


class hipErrorAlreadyAcquired(hipError):
    __doc__ = _libhip.hipGetErrorString(210)
    pass


class hipErrorNotMapped(hipError):
    __doc__ = _libhip.hipGetErrorString(211)
    pass


class hipErrorNotMappedAsArray(hipError):
    __doc__ = _libhip.hipGetErrorString(212)
    pass


class hipErrorNotMappedAsPointer(hipError):
    __doc__ = _libhip.hipGetErrorString(213)
    pass


class hipErrorECCNotCorrectable(hipError):
    __doc__ = _libhip.hipGetErrorString(214)
    pass


class hipErrorUnsupportedLimit(hipError):
    __doc__ = _libhip.hipGetErrorString(215)
    pass


class hipErrorContextAlreadyInUse(hipError):
    __doc__ = _libhip.hipGetErrorString(216)
    pass


class hipErrorPeerAccessUnsupported(hipError):
    __doc__ = _libhip.hipGetErrorString(217)
    pass


class hipErrorInvalidKernelFile(hipError):
    __doc__ = _libhip.hipGetErrorString(218)
    pass


class hipErrorInvalidGraphicsContext(hipError):
    __doc__ = _libhip.hipGetErrorString(219)
    pass


class hipErrorInvalidSource(hipError):
    __doc__ = _libhip.hipGetErrorString(300)
    pass


class hipErrorFileNotFound(hipError):
    __doc__ = _libhip.hipGetErrorString(301)
    pass


class hipErrorSharedObjectSymbolNotFound(hipError):
    __doc__ = _libhip.hipGetErrorString(302)
    pass


class hipErrorSharedObjectInitFailed(hipError):
    __doc__ = _libhip.hipGetErrorString(303)
    pass


class hipErrorOperatingSystem(hipError):
    __doc__ = _libhip.hipGetErrorString(304)
    pass


class hipErrorInvalidHandle(hipError):
    __doc__ = _libhip.hipGetErrorString(400)
    pass


class hipErrorNotFound(hipError):
    __doc__ = _libhip.hipGetErrorString(500)
    pass


class hipErrorNotReady(hipError):
    __doc__ = _libhip.hipGetErrorString(600)
    pass


class hipErrorIllegalAddress(hipError):
    __doc__ = _libhip.hipGetErrorString(700)
    pass


class hipErrorLaunchOutOfResources(hipError):
    __doc__ = _libhip.hipGetErrorString(701)
    pass


class hipErrorLaunchTimeOut(hipError):
    __doc__ = _libhip.hipGetErrorString(702)
    pass


class hipErrorPeerAccessAlreadyEnabled(hipError):
    __doc__ = _libhip.hipGetErrorString(704)
    pass


class hipErrorPeerAccessNotEnabled(hipError):
    __doc__ = _libhip.hipGetErrorString(705)
    pass


class hipErrorSetOnActiveProcess(hipError):
    __doc__ = _libhip.hipGetErrorString(708)
    pass


class hipErrorAssert(hipError):
    __doc__ = _libhip.hipGetErrorString(710)
    pass


class hipErrorHostMemoryAlreadyRegistered(hipError):
    __doc__ = _libhip.hipGetErrorString(712)
    pass


class hipErrorHostMemoryNotRegistered(hipError):
    __doc__ = _libhip.hipGetErrorString(713)
    pass


class hipErrorLaunchFailure(hipError):
    __doc__ = _libhip.hipGetErrorString(719)
    pass


class hipErrorCooperativeLaunchTooLarge(hipError):
    __doc__ = _libhip.hipGetErrorString(720)
    pass


class hipErrorNotSupported(hipError):
    __doc__ = _libhip.hipGetErrorString(801)
    pass


class hipErrorUnknown(hipError):
    __doc__ = _libhip.hipGetErrorString(999)
    pass


class hipErrorRuntimeMemory(hipError):
    __doc__ = _libhip.hipGetErrorString(1052)
    pass


class hipErrorRuntimeOther(hipError):
    __doc__ = _libhip.hipGetErrorString(1053)
    pass


hipExceptions = {
    1: hipErrorInvalidValue,
    2: hipErrorOutOfMemory,
    3: hipErrorNotInitialized,
    4: hipErrorDeinitialized,
    5: hipErrorProfilerDisabled,
    6: hipErrorProfilerNotInitialized,
    7: hipErrorProfilerAlreadyStarted,
    8: hipErrorProfilerAlreadyStopped,
    9: hipErrorInvalidConfiguration,
    13: hipErrorInvalidSymbol,
    17: hipErrorInvalidDevicePointer,
    21: hipErrorInvalidMemcpyDirection,
    35: hipErrorInsufficientDriver,
    52: hipErrorMissingConfiguration,
    53: hipErrorPriorLaunchFailure,
    98: hipErrorInvalidDeviceFunction,
    100: hipErrorNoDevice,
    101: hipErrorInvalidDevice,
    200: hipErrorInvalidImage,
    201: hipErrorInvalidContext,
    202: hipErrorContextAlreadyCurrent,
    205: hipErrorMapFailed,
    206: hipErrorUnmapFailed,
    207: hipErrorArrayIsMapped,
    208: hipErrorAlreadyMapped,
    209: hipErrorNoBinaryForGpu,
    210: hipErrorAlreadyAcquired,
    211: hipErrorNotMapped,
    212: hipErrorNotMappedAsArray,
    213: hipErrorNotMappedAsPointer,
    214: hipErrorECCNotCorrectable,
    215: hipErrorUnsupportedLimit,
    216: hipErrorContextAlreadyInUse,
    217: hipErrorPeerAccessUnsupported,
    218: hipErrorInvalidKernelFile,
    219: hipErrorInvalidGraphicsContext,
    300: hipErrorInvalidSource,
    301: hipErrorFileNotFound,
    302: hipErrorSharedObjectSymbolNotFound,
    303: hipErrorSharedObjectInitFailed,
    304: hipErrorOperatingSystem,
    400: hipErrorInvalidHandle,
    500: hipErrorNotFound,
    600: hipErrorNotReady,
    700: hipErrorIllegalAddress,
    701: hipErrorLaunchOutOfResources,
    702: hipErrorLaunchTimeOut,
    704: hipErrorPeerAccessAlreadyEnabled,
    705: hipErrorPeerAccessNotEnabled,
    708: hipErrorSetOnActiveProcess,
    710: hipErrorAssert,
    712: hipErrorHostMemoryAlreadyRegistered,
    713: hipErrorHostMemoryNotRegistered,
    719: hipErrorLaunchFailure,
    720: hipErrorCooperativeLaunchTooLarge,
    801: hipErrorNotSupported,
    999: hipErrorUnknown,
    1052: hipErrorRuntimeMemory,
    1053: hipErrorRuntimeOther
}


def hipCheckStatus(status):
    """
    Raise hip exception.

    Raise an exception corresponding to the specified hip runtime error
    code.

    Parameters
    ----------
    status : int
        hip runtime error code.

    See Also
    --------
    hipExceptions
    """

    if status != 0:
        try:
            e = hipExceptions[status]
        except KeyError:
            raise hipError('unknown hip error %s' % status)
        else:
            raise e


# Stream management

_libhip.hipStreamCreate.restype = int
_libhip.hipStreamCreate.argtypes = [ctypes.POINTER(ctypes.c_void_p)]


def hipStreamCreate():
    """
    Create an asynchronous stream.

    Create a new asynchronous stream.  @p stream returns an opaque handle that can be used to
    reference the newly created stream in subsequent hipStream* commands.  The stream is allocated on
    the heap and will remain allocated even if the handle goes out-of-scope.  To release the memory
    used by the stream, application must call hipStreamDestroy.

    Returns
    -------
    ptr : ctypes pointer
        Valid pointer to stream object.
    """
    ptr = ctypes.c_void_p()
    status = _libhip.hipStreamCreate(ctypes.byref(ptr))
    hipCheckStatus(status)
    return ptr


_libhip.hipStreamDestroy.restype = int
_libhip.hipStreamDestroy.argtypes = [ctypes.c_void_p]


def hipStreamDestroy(ptr):
    """
    Destroys the specified stream.
    
    If commands are still executing on the specified stream, some may complete execution before the
    queue is deleted.
    
    The queue may be destroyed while some commands are still inflight, or may wait for all commands
    queued to the stream before destroying it.

    Parameters
    -------
    ptr : ctypes pointer
        Valid pointer to stream object.
    """
    status = _libhip.hipStreamDestroy(ptr)
    hipCheckStatus(status)


_libhip.hipStreamDestroy.restype = int
_libhip.hipStreamDestroy.argtypes = [ctypes.c_void_p]


def hipStreamDestroy(ptr):
    """
    Destroys the specified stream.
    
    If commands are still executing on the specified stream, some may complete execution before the
    queue is deleted.
    
    The queue may be destroyed while some commands are still inflight, or may wait for all commands
    queued to the stream before destroying it.

    Parameters
    -------
    ptr : ctypes pointer
        Valid pointer to stream object.
    """
    status = _libhip.hipStreamDestroy(ptr)
    hipCheckStatus(status)


_libhip.hipStreamSynchronize.restype = int
_libhip.hipStreamSynchronize.argtypes = [ctypes.c_void_p]


def hipStreamSynchronize(ptr):
    """
    Wait for all commands in stream to complete.
    
    This command is host-synchronous : the host will block until the specified stream is empty.
    
    This command follows standard null-stream semantics.  Specifically, specifying the null stream
    will cause the command to wait for other streams on the same device to complete all pending
    operations.
    
    This command honors the hipDeviceLaunchBlocking flag, which controls whether the wait is active
    or blocking.

    Parameters
    -------
    ptr : ctypes pointer
        Valid pointer to stream object.
    """
    status = _libhip.hipStreamSynchronize(ptr)
    hipCheckStatus(status)


# Event management

# Event creation flags
hipEventDefault = 0
hipEventBlockingSync = 1
hipEventDisableTiming = 2
hipEventInterprocess = 4

_libhip.hipEventCreateWithFlags.restype = int
_libhip.hipEventCreateWithFlags.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_int]


def hipEventCreateWithFlags(flags):
    """
    Create an event with the specified flags

    Parameters
    ----------
    flags : int
        Flags to control event behavior.  Valid values are hipEventDefault,
        hipEventBlockingSync, hipEventDisableTiming, #hipEventInterprocess
        * hipEventDefault : Default flag.  The event will use active synchronization and will support
        timing.  Blocking synchronization provides lowest possible latency at the expense of dedicating a
        CPU to poll on the event.
        * hipEventBlockingSync : The event will use blocking synchronization : if hipEventSynchronize is
        called on this event, the thread will block until the event completes.  This can increase latency
        for the synchroniation but can result in lower power and more resources for other CPU threads.
        * hipEventDisableTiming : Disable recording of timing information. Events created with this flag
        would not record profiling data and provide best performance if used for synchronization.
        * hipEventInterprocess : Warning On AMD platform, support is under development.  Use of this flag
        will return an error.

    Returns
    -------
    ptr : ctypes pointer
        Pointer to the newly created event.

    """
    ptr = ctypes.c_void_p()
    status = _libhip.hipEventCreateWithFlags(ctypes.byref(ptr), flags)
    hipCheckStatus(status)
    return ptr


_libhip.hipEventCreate.restype = int
_libhip.hipEventCreate.argtypes = [ctypes.POINTER(ctypes.c_void_p)]


def hipEventCreate():
    """
    Create an event with the specified flags

    Returns
    -------
    ptr : ctypes pointer
        Pointer to the newly created event.

    """
    ptr = ctypes.c_void_p()
    status = _libhip.hipEventCreate(ctypes.byref(ptr))
    hipCheckStatus(status)
    return ptr


_libhip.hipEventRecord.restype = int
_libhip.hipEventRecord.argtypes = [ctypes.c_void_p, ctypes.c_void_p]


def hipEventRecord(event, stream=None):
    """
    Record an event in the specified stream.

    Parameters
    ----------
    event : handle
        event to record.
    stream : _ctypes pointer, optional
        stream in which to record event.

    """
    status = _libhip.hipEventRecord(event, stream)
    hipCheckStatus(status)


_libhip.hipEventDestroy.restype = int
_libhip.hipEventDestroy.argtypes = [ctypes.c_void_p]


def hipEventDestroy(ptr):
    """
    Destroy the specified event.

    Parameters
    ----------
    event : ctypes pointer
        Event to destroy.

    """
    status = _libhip.hipEventDestroy(ptr)
    hipCheckStatus(status)


_libhip.hipEventSynchronize.restype = int
_libhip.hipEventSynchronize.argtypes = [ctypes.c_void_p]


def hipEventSynchronize(ptr):
    """
    Wait for an event to complete.

    This function will block until the event is ready, waiting for all previous work in the stream
    specified when event was recorded with hipEventRecord().

    If hipEventRecord() has not been called on @p event, this function returns immediately.

    Parameters
    ----------
    event : ctypes pointer
        Event to Synchronize.

    """
    status = _libhip.hipEventSynchronize(ptr)
    hipCheckStatus(status)


_libhip.hipEventElapsedTime.restype = int
_libhip.hipEventElapsedTime.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.c_void_p, ctypes.c_void_p]


def hipEventElapsedTime(start, stop):
    """
    Return the elapsed time between two events.

    Computes the elapsed time between two events. Time is computed in ms, with
    a resolution of approximately 1 us.
    
    Events which are recorded in a NULL stream will block until all commands
    on all other streams complete execution, and then record the timestamp.
    
    Events which are recorded in a non-NULL stream will record their timestamp
    when they reach the head of the specified stream, after all previous
    commands in that stream have completed executing.  Thus the time that
    the event recorded may be significantly after the host calls hipEventRecord().
    Parameters
    ----------
    start : ctypes pointer
        Start event.
    stop : ctypes pointer
        Stop event.
    """
    t = ctypes.c_float()
    status = _libhip.hipEventElapsedTime(ctypes.byref(t), start, stop)
    hipCheckStatus(status)
    return t.value


# Memory allocation functions (adapted from pystream):
_libhip.hipMalloc.restype = int
_libhip.hipMalloc.argtypes = [ctypes.POINTER(ctypes.c_void_p), ctypes.c_size_t]


def hipMalloc(count, ctype=None):
    """
    Allocate device memory.

    Allocate memory on the device associated with the current active
    context.

    Parameters
    ----------
    count : int
        Number of bytes of memory to allocate
    ctype : _ctypes.SimpleType, optional
        ctypes type to cast returned pointer.

    Returns
    -------
    ptr : ctypes pointer
        Pointer to allocated device memory.

    """

    ptr = ctypes.c_void_p()
    status = _libhip.hipMalloc(ctypes.byref(ptr), count)
    hipCheckStatus(status)
    if ctype is not None:
        ptr = ctypes.cast(ptr, ctypes.POINTER(ctype))
    return ptr


_libhip.hipFree.restype = int
_libhip.hipFree.argtypes = [ctypes.c_void_p]


def hipFree(ptr):
    """
    Free device memory.

    Free allocated memory on the device associated with the current active
    context.

    Parameters
    ----------
    ptr : ctypes pointer
        Pointer to allocated device memory.

    """

    status = _libhip.hipFree(ptr)
    hipCheckStatus(status)


_libhip.hipMallocPitch.restype = int
_libhip.hipMallocPitch.argtypes = [
    ctypes.POINTER(ctypes.c_void_p),
    ctypes.POINTER(ctypes.c_size_t), ctypes.c_size_t, ctypes.c_size_t
]


def hipMallocPitch(pitch, rows, cols, elesize):
    """
    Allocate pitched device memory.

    Allocate pitched memory on the device associated with the current active
    context.

    Parameters
    ----------
    pitch : int
        Pitch for allocation.
    rows : int
        Requested pitched allocation height.
    cols : int
        Requested pitched allocation width.
    elesize : int
        Size of memory element.

    Returns
    -------
    ptr : ctypes pointer
        Pointer to allocated device memory.

    """

    ptr = ctypes.c_void_p()
    status = _libhip.hipMallocPitch(ctypes.byref(ptr), ctypes.c_size_t(pitch), cols * elesize, rows)
    hipCheckStatus(status)
    return ptr, pitch


# Memory copy modes:
hipMemcpyHostToHost = 0
hipMemcpyHostToDevice = 1
hipMemcpyDeviceToHost = 2
hipMemcpyDeviceToDevice = 3
hipMemcpyDefault = 4

_libhip.hipMemcpy.restype = int
_libhip.hipMemcpy.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_int]


def hipMemcpy_htod(dst, src, count):
    """
    Copy memory from host to device.

    Copy data from host memory to device memory.

    Parameters
    ----------
    dst : ctypes pointer
        Device memory pointer.
    src : ctypes pointer
        Host memory pointer.
    count : int
        Number of bytes to copy.

    """

    status = _libhip.hipMemcpy(dst, src, ctypes.c_size_t(count), hipMemcpyHostToDevice)
    hipCheckStatus(status)


def hipMemcpy_dtoh(dst, src, count):
    """
    Copy memory from device to host.

    Copy data from device memory to host memory.

    Parameters
    ----------
    dst : ctypes pointer
        Host memory pointer.
    src : ctypes pointer
        Device memory pointer.
    count : int
        Number of bytes to copy.

    """

    status = _libhip.hipMemcpy(dst, src, ctypes.c_size_t(count), hipMemcpyDeviceToHost)
    hipCheckStatus(status)


_libhip.hipMemGetInfo.restype = int
_libhip.hipMemGetInfo.argtypes = [ctypes.c_void_p, ctypes.c_void_p]


def hipMemGetInfo():
    """
    Return the amount of free and total device memory.

    Returns
    -------
    free : long
        Free memory in bytes.
    total : long
        Total memory in bytes.

    """

    free = ctypes.c_size_t()
    total = ctypes.c_size_t()
    status = _libhip.hipMemGetInfo(ctypes.byref(free), ctypes.byref(total))
    hipCheckStatus(status)
    return free.value, total.value


_libhip.hipSetDevice.restype = int
_libhip.hipSetDevice.argtypes = [ctypes.c_int]


def hipSetDevice(dev):
    """
    Set current hip device.

    Select a device to use for subsequent hip operations.

    Parameters
    ----------
    dev : int
        Device number.

    """

    status = _libhip.hipSetDevice(dev)
    hipCheckStatus(status)


_libhip.hipGetDevice.restype = int
_libhip.hipGetDevice.argtypes = [ctypes.POINTER(ctypes.c_int)]


def hipGetDevice():
    """
    Get current hip device.

    Return the identifying number of the device currently used to
    process hip operations.

    Returns
    -------
    dev : int
        Device number.

    """

    dev = ctypes.c_int()
    status = _libhip.hipGetDevice(ctypes.byref(dev))
    hipCheckStatus(status)
    return dev.value


class hipDeviceArch(ctypes.Structure):
    _fields_ = [
    # *32-bit Atomics*
    # 32-bit integer atomics for global memory.
        ('hasGlobalInt32Atomics', ctypes.c_uint, 1),

    # 32-bit float atomic exch for global memory.
        ('hasGlobalFloatAtomicExch', ctypes.c_uint, 1),

    # 32-bit integer atomics for shared memory.
        ('hasSharedInt32Atomics', ctypes.c_uint, 1),

    # 32-bit float atomic exch for shared memory.
        ('hasSharedFloatAtomicExch', ctypes.c_uint, 1),

    # 32-bit float atomic add in global and shared memory.
        ('hasFloatAtomicAdd', ctypes.c_uint, 1),

    # *64-bit Atomics*
    # 64-bit integer atomics for global memory.
        ('hasGlobalInt64Atomics', ctypes.c_uint, 1),

    # 64-bit integer atomics for shared memory.
        ('hasSharedInt64Atomics', ctypes.c_uint, 1),

    # *Doubles*
    # Double-precision floating point.
        ('hasDoubles', ctypes.c_uint, 1),

    # *Warp cross-lane operations*
    # Warp vote instructions (__any, __all).
        ('hasWarpVote', ctypes.c_uint, 1),

    # Warp ballot instructions (__ballot).
        ('hasWarpBallot', ctypes.c_uint, 1),

    # Warp shuffle operations. (__shfl_*).
        ('hasWarpShuffle', ctypes.c_uint, 1),

    # Funnel two words into one with shift&mask caps.
        ('hasFunnelShift', ctypes.c_uint, 1),

    # *Sync*
    # __threadfence_system.
        ('hasThreadFenceSystem', ctypes.c_uint, 1),

    # __syncthreads_count, syncthreads_and, syncthreads_or.
        ('hasSyncThreadsExt', ctypes.c_uint, 1),

    # *Misc*
    # Surface functions.
        ('hasSurfaceFuncs', ctypes.c_uint, 1),

    # Grid and group dims are 3D (rather than 2D).
        ('has3dGrid', ctypes.c_uint, 1),

    # Dynamic parallelism.
        ('hasDynamicParallelism', ctypes.c_uint, 1),
    ]


class hipDeviceProperties(ctypes.Structure):
    _fields_ = [
    # Device name
        ('_name', ctypes.c_char * 256),

    # Size of global memory region (in bytes)
        ('totalGlobalMem', ctypes.c_size_t),

    # Size of shared memory region (in bytes).
        ('sharedMemPerBlock', ctypes.c_size_t),

    # Registers per block.
        ('regsPerBlock', ctypes.c_int),

    # Warp size.
        ('warpSize', ctypes.c_int),

    # Max work items per work group or workgroup max size.
        ('maxThreadsPerBlock', ctypes.c_int),

    # Max number of threads in each dimension (XYZ) of a block.
        ('maxThreadsDim', ctypes.c_int * 3),

    # Max grid dimensions (XYZ).
        ('maxGridSize', ctypes.c_int * 3),

    # Max clock frequency of the multiProcessors in khz.
        ('clockRate', ctypes.c_int),

    # Max global memory clock frequency in khz.
        ('memoryClockRate', ctypes.c_int),

    # Global memory bus width in bits.
        ('memoryBusWidth', ctypes.c_int),

    # Size of shared memory region (in bytes).
        ('totalConstMem', ctypes.c_size_t),

    # Major compute capability.  On HCC, this is an approximation and features may
    # differ from CUDA CC.  See the arch feature flags for portable ways to query
    # feature caps.
        ('major', ctypes.c_int),

    # Minor compute capability.  On HCC, this is an approximation and features may
    # differ from CUDA CC.  See the arch feature flags for portable ways to query
    # feature caps.
        ('minor', ctypes.c_int),

    # Number of multi-processors (compute units).
        ('multiProcessorCount', ctypes.c_int),

    # L2 cache size.
        ('l2CacheSize', ctypes.c_int),

    # Maximum resident threads per multi-processor.
        ('maxThreadsPerMultiProcessor', ctypes.c_int),

    # Compute mode.
        ('computeMode', ctypes.c_int),

    # Frequency in khz of the timer used by the device-side "clock*"
    # instructions.  New for HIP.
        ('clockInstructionRate', ctypes.c_int),

    # Architectural feature flags.  New for HIP.
        ('arch', hipDeviceArch),

    # Device can possibly execute multiple kernels concurrently.
        ('concurrentKernels', ctypes.c_int),

    # PCI Domain ID
        ('pciDomainID', ctypes.c_int),

    # PCI Bus ID.
        ('pciBusID', ctypes.c_int),

    # PCI Device ID.
        ('pciDeviceID', ctypes.c_int),

    # Maximum Shared Memory Per Multiprocessor.
        ('maxSharedMemoryPerMultiProcessor', ctypes.c_size_t),

    # 1 if device is on a multi-GPU board, 0 if not.
        ('isMultiGpuBoard', ctypes.c_int),

    # Check whether HIP can map host memory
        ('canMapHostMemory', ctypes.c_int),

    # DEPRECATED: use gcnArchName instead
        ('gcnArch', ctypes.c_int),

    # AMD GCN Arch Name.
        ('_gcnArchName', ctypes.c_char * 256),

    # APU vs dGPU
        ('integrated', ctypes.c_int),

    # HIP device supports cooperative launch
        ('cooperativeLaunch', ctypes.c_int),

    # HIP device supports cooperative launch on multiple devices
        ('cooperativeMultiDeviceLaunch', ctypes.c_int),

    # Maximum size for 1D textures bound to linear memory
        ('maxTexture1DLinear', ctypes.c_int),

    # Maximum number of elements in 1D images
        ('maxTexture1D', ctypes.c_int),

    # Maximum dimensions (width, height) of 2D images, in image elements
        ('maxTexture2D', ctypes.c_int * 2),

    # Maximum dimensions (width, height, depth) of 3D images, in image elements
        ('maxTexture3D', ctypes.c_int * 3),

    # Addres of HDP_MEM_COHERENCY_FLUSH_CNTL register
        ('hdpMemFlushCntl', POINTER(ctypes.c_uint)),

    # Addres of HDP_REG_COHERENCY_FLUSH_CNTL register
        ('hdpRegFlushCntl', POINTER(ctypes.c_uint)),

    # Maximum pitch in bytes allowed by memory copies
        ('memPitch', ctypes.c_size_t),

    # Alignment requirement for textures
        ('textureAlignment', ctypes.c_size_t),

    # Pitch alignment requirement for texture references bound to pitched memory
        ('texturePitchAlignment', ctypes.c_size_t),

    # Run time limit for kernels executed on the device
        ('kernelExecTimeoutEnabled', ctypes.c_int),

    # Device has ECC support enabled
        ('ECCEnabled', ctypes.c_int),

    # 1:If device is Tesla device using TCC driver, else 0
        ('tccDriver', ctypes.c_int),

    # HIP device supports cooperative launch on multiple
    # devices with unmatched functions
        ('cooperativeMultiDeviceUnmatchedFunc', ctypes.c_int),

    # HIP device supports cooperative launch on multiple
    # devices with unmatched grid dimensions
        ('cooperativeMultiDeviceUnmatchedGridDim', ctypes.c_int),

    # HIP device supports cooperative launch on multiple
    # devices with unmatched block dimensions
        ('cooperativeMultiDeviceUnmatchedBlockDim', ctypes.c_int),

    # HIP device supports cooperative launch on multiple
    # devices with unmatched shared memories
        ('cooperativeMultiDeviceUnmatchedSharedMem', ctypes.c_int),

    # 1: if it is a large PCI bar device, else 0
        ('isLargeBar', ctypes.c_int),

    # Revision of the GPU in this device
        ('asicRevision', ctypes.c_int),

    # Device supports allocating managed memory on this system
        ('managedMemory', ctypes.c_int),

    # Host can directly access managed memory on the device without migration
        ('directManagedMemAccessFromHost', ctypes.c_int),

    # Device can coherently access managed memory concurrently with the CPU
        ('concurrentManagedAccess', ctypes.c_int),

    # Device supports coherently accessing pageable memory
    # without calling hipHostRegister on it
        ('pageableMemoryAccess', ctypes.c_int),

    # Device accesses pageable memory via the host's page tables
        ('pageableMemoryAccessUsesHostPageTables', ctypes.c_int),
    ]

    @property
    def name(self):
        return self._name.decode('utf-8')

    @property
    def gcnArchName(self):
        return self._gcnArchName.decode('utf-8')


_libhip.hipGetDeviceProperties.restype = int
_libhip.hipGetDeviceProperties.argtypes = [POINTER(hipDeviceProperties), ctypes.c_int]


def hipGetDeviceProperties(deviceId: int):
    """
    Populates hipGetDeviceProperties with information for the specified device.
    Returns device properties of the specified device.
    Parameters
    ----------
    deviceId : int
        Which device to query for information
    Returns
    -------
    device_properties : hipDeviceProperties
       Information for the specified device
    """
    device_properties = hipDeviceProperties()
    status = _libhip.hipGetDeviceProperties(ctypes.pointer(device_properties), deviceId)
    hipCheckStatus(status)
    return device_properties


# Memory types:
hipMemoryTypeHost = 1
hipMemoryTypeDevice = 2
hipMemoryTypeArray = 3
hipMemoryTypeUnified = 4    # Not used currently


class hipPointerAttributes(ctypes.Structure):
    _fields_ = [('memoryType', ctypes.c_int), ('device', ctypes.c_int), ('devicePointer', ctypes.c_void_p),
                ('hostPointer', ctypes.c_void_p)]


_libhip.hipPointerGetAttributes.restype = int
_libhip.hipPointerGetAttributes.argtypes = [ctypes.c_void_p, ctypes.c_void_p]


def hipPointerGetAttributes(ptr):
    """
    Get memory pointer attributes.

    Returns attributes of the specified pointer.

    Parameters
    ----------
    ptr : ctypes pointer
        Memory pointer to examine.

    Returns
    -------
    memory_type : int
        Memory type; 1 indicates host memory, 2 indicates device
        memory.
    device : int
        Number of device associated with pointer.
    """

    attributes = hipPointerAttributes()
    status = \
        _libhip.hipPointerGetAttributes(ctypes.byref(attributes), ptr)
    hipCheckStatus(status)
    return attributes.memoryType, attributes.device


_libhip.hipModuleLoadData.restype = int
_libhip.hipModuleLoadData.argtypes = [
    ctypes.POINTER(ctypes.c_void_p),    # Module
    ctypes.c_void_p
]    # Image


def hipModuleLoadData(data):
    """
    Load hip module data

    Parameters
    ----------
    data : ctypes pointer
        Memory pointer to load.

    Returns
    -------
    module : ctypes ptr
        hip module
    """

    module = ctypes.c_void_p()
    status = _libhip.hipModuleLoadData(ctypes.byref(module), data)
    hipCheckStatus(status)
    return module


_libhip.hipModuleGetFunction.restype = int
_libhip.hipModuleGetFunction.argtypes = [
    ctypes.POINTER(ctypes.c_void_p),    # Kernel
    ctypes.c_void_p,    # Module
    ctypes.POINTER(ctypes.c_char)
]    # kernel name


def hipModuleGetFunction(module, func_name):
    """
    gets the kernel from module

    Parameters
    ----------
    module : ctypes ptr
        pointer to created module
    func_name : string
        name of function to be retrived from module

    Returns
    -------
    kernel : ctypes ptr
        pointer to kernel type
    """
    e_func_name = func_name.encode('utf-8')
    kernel = ctypes.c_void_p()
    status = _libhip.hipModuleGetFunction(ctypes.byref(kernel), module, e_func_name)
    hipCheckStatus(status)
    return kernel


_libhip.hipModuleUnload.restype = int
_libhip.hipModuleUnload.argtypes = [ctypes.c_void_p]


def hipModuleUnload(module):
    """
    gets the kernel from module

    Parameters
    ----------
    module : ctypes ptr
        pointer to created module
    """
    status = _libhip.hipModuleUnload(module)
    hipCheckStatus(status)


_libhip.hipModuleLaunchKernel.restype = int
_libhip.hipModuleLaunchKernel.argtypes = [
    ctypes.c_void_p,    # kernel
    ctypes.c_uint,    # block x
    ctypes.c_uint,    # block y
    ctypes.c_uint,    # block z
    ctypes.c_uint,    # thread x
    ctypes.c_uint,    # thread y
    ctypes.c_uint,    # thread z
    ctypes.c_uint,    # shared mem
    ctypes.c_void_p,    # stream
    ctypes.POINTER(ctypes.c_void_p),    # kernel params
    ctypes.POINTER(ctypes.c_void_p)
]    # extra


def hipModuleLaunchKernel(kernel, bx, by, bz, tx, ty, tz, shared, stream, struct):
    """
    Launch the kernel

    Parameters
    ----------
    kernel : ctypes ptr
        kernel from loaded module
    bx : int
        dim x
    by : int
        dim y
    bz : int
        dim z
    tx : int
        dim x
    ty : int
        dim y
    tz : int
        dim z
    shared : int
        shared mem
    stream : ctype void ptr
        stream object
    struct : ctypes structure
        struct of packed up arguments of kernel
    """
    c_bx = ctypes.c_uint(bx)
    c_by = ctypes.c_uint(by)
    c_bz = ctypes.c_uint(bz)
    c_tx = ctypes.c_uint(tx)
    c_ty = ctypes.c_uint(ty)
    c_tz = ctypes.c_uint(tz)
    c_shared = ctypes.c_uint(shared)

    ctypes.sizeof(struct)
    hip_launch_param_buffer_ptr = ctypes.c_void_p(1)
    hip_launch_param_buffer_size = ctypes.c_void_p(2)
    hip_launch_param_buffer_end = ctypes.c_void_p(3)
    size = ctypes.c_size_t(ctypes.sizeof(struct))
    p_size = ctypes.c_void_p(ctypes.addressof(size))
    p_struct = ctypes.c_void_p(ctypes.addressof(struct))
    config = (ctypes.c_void_p * 5)(
        hip_launch_param_buffer_ptr, p_struct, hip_launch_param_buffer_size, p_size, hip_launch_param_buffer_end
    )
    nullptr = ctypes.POINTER(ctypes.c_void_p)(ctypes.c_void_p(0))

    status = _libhip.hipModuleLaunchKernel(kernel, c_bx, c_by, c_bz, c_tx, c_ty, c_tz, c_shared, stream, None, config)
    hipCheckStatus(status)


_libhip.hipDeviceSynchronize.restype = int
_libhip.hipDeviceSynchronize.argtypes = []


def hipDeviceSynchronize():
    """
    Device level sync
    """
    status = _libhip.hipDeviceSynchronize()
    hipCheckStatus(status)


_libhip.hipInit.restype = int
_libhip.hipInit.argtypes = [ctypes.c_uint]    # flags


def hipInit(flags):
    """
    Explicitly initializes the HIP runtime.

    Most HIP APIs implicitly initialize the HIP runtime.
    This API provides control over the timing of the initialization.

    Parameters
    ----------
    flags : int
    """
    c_flags = ctypes.c_uint(flags)
    status = _libhip.hipInit(c_flags)
    hipCheckStatus(status)



#!/usr/bin/env python3

"""Converts a HAT package with .obj/.o into a HAT package with a .lib/.a

A statically-linked HAT package contains a binary file with the extension '.o', '.obj', '.a', or `.lib`.

This tool converts a statically-linked HAT package with an '.o' or '.obj' binary file into another
statically-linked HAT package with a '.a' or '.lib' binary file.

To use the tool, point it to the '.hat' file associated with the statically-linked package (the
'.hat' file knows where to find the associated binary file), and provide a filename for the new
'.hat' file.

Dependencies on Windows:
* the cl.exe command-line compiler, available with Microsoft Visual Studio

Dependencies on Linux / macOS:
* the gcc command-line compiler
"""

import sys
import os
import argparse
import shutil

from .hat_file import HATFile, OperatingSystem
from .platform_utilities import get_platform, ensure_compiler_in_path, run_command


def linux_create_static_package(input_hat_binary_path, output_hat_path, hat_file, quiet=True):
    """Creates a static HAT (.a) from a static HAT (.o) on a Linux/macOS platform"""
    # Confirm that this is a static .o hat library
    _, extension = os.path.splitext(input_hat_binary_path)
    if extension != ".o":
        sys.exit(f"ERROR: Expected input library to have extension .o, but received {input_hat_binary_path} instead")

    # create new HAT binary
    prefix, _ = os.path.splitext(output_hat_path)
    output_hat_binary_path = f"{prefix}.a"
    libraries = " ".join([d.target_file for d in hat_file.dependencies.dynamic])
    run_command(f'ar rcs "{output_hat_binary_path}" "{input_hat_binary_path}" {libraries}', quiet=quiet)

    # create new HAT file
    hat_file.dependencies.dynamic = [] # previous dependencies are now part of the binary
    hat_file.dependencies.link_target = os.path.basename(output_hat_binary_path)
    hat_file.Serialize(output_hat_path)


def windows_create_static_package(input_hat_binary_path, output_hat_path, hat_file, quiet=True):
    """Creates a Windows static .lib HAT package from a static .obj HAT package"""

    # Confirm that this is a static hat library
    _, extension = os.path.splitext(input_hat_binary_path)
    if extension != ".obj":
        sys.exit(f"ERROR: Expected input library to have extension .obj, but received {input_hat_binary_path} instead")

    # Create all file in a directory named build
    if not os.path.exists("build"):
        os.mkdir("build")

    cwd = os.getcwd()
    try:
        os.chdir("build")

        # create the new HAT binary lib
        prefix, _ = os.path.splitext(output_hat_path)
        output_hat_binary_path = f"{prefix}.lib"

        # presume /DEF is not needed because the exports will be part of the HAT file
        libraries = " ".join([d.target_file for d in hat_file.dependencies.dynamic])
        archiver_command_line = f'lib.exe /NOLOGO /OUT:out.lib "{input_hat_binary_path}" {libraries}'
        run_command(archiver_command_line, quiet=quiet)
        shutil.copyfile("out.lib", output_hat_binary_path)

        # create new HAT file
        hat_file.dependencies.dynamic = [] # previous dependencies are now part of the binary
        hat_file.dependencies.link_target = os.path.basename(output_hat_binary_path)
        hat_file.Serialize("out.hat")
        shutil.copyfile("out.hat", output_hat_path)
    finally:
        os.chdir(cwd) # restore the current working directory

def parse_args():
    """Parses and checks the command line arguments"""
    parser = argparse.ArgumentParser(description="Creates a statically-linked HAT package with a .lib/.a from a statically-linked HAT package with an .obj/.o.\n"
        "Example:\n"
        "    python hatlib.hat_to_lib input.hat output.hat")

    parser.add_argument("input_hat_path", type=str, help="Path to the existing HAT file, which represents a statically-linked HAT package with a .obj or .o binary file")
    parser.add_argument("output_hat_path", type=str, help="Path to the new HAT file, which will represent a statically-linked HAT package with a .lib or .a binary file")
    parser.add_argument('-v', "--verbose", action='store_true', help="Enable verbose output")
    args = parser.parse_args()

    # check args
    if not os.path.exists(args.input_hat_path):
        sys.exit(f"ERROR: File {args.input_hat_path} not found")

    if os.path.abspath(args.input_hat_path) == os.path.abspath(args.output_hat_path):
        sys.exit("ERROR: Output file must be different from input file")

    _, extension = os.path.splitext(args.input_hat_path)
    if extension != ".hat":
        sys.exit(f"ERROR: Expected input file to have extension .hat, but received {extension} instead")

    _, extension = os.path.splitext(args.output_hat_path)
    if extension != ".hat":
        sys.exit(f"ERROR: Expected output file to have extension .hat, but received {extension} instead")

    return args


def create_static_package(input_hat_path, output_hat_path, quiet=True):
    platform = get_platform()
    ensure_compiler_in_path()

    # load the function decscriptions and the library path from the hat file
    input_hat_path = os.path.abspath(input_hat_path)
    hat_file = HATFile.Deserialize(input_hat_path)

    # get the absolute path to the input binary
    input_hat_binary_filename = hat_file.dependencies.link_target
    input_hat_binary_path = os.path.join(os.path.dirname(input_hat_path), input_hat_binary_filename)

    # create the static library package
    # TODO: prefer lld when available and support cross-compilation
    output_hat_path = os.path.abspath(output_hat_path)
    if platform == OperatingSystem.Windows:
        windows_create_static_package(input_hat_binary_path, output_hat_path, hat_file, quiet=quiet)
    elif platform in [OperatingSystem.Linux, OperatingSystem.MacOS]:
        linux_create_static_package(input_hat_binary_path, output_hat_path, hat_file, quiet=quiet)


def main():
    args = parse_args()
    create_static_package(args.input_hat_path, args.output_hat_path, quiet=not args.verbose)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Utility to parse and validate a HAT package

import ctypes
from typing import Any, List, Union
from collections import OrderedDict
from functools import partial

from .hat_file import HATFile, Function, Parameter
from .arg_info import ArgInfo, verify_args

import os


class HATPackage:

    def __init__(self, hat_file_path):
        """A HAT Package is defined to be a HAT file and corresponding binary file, located in the same directory.
        The binary file is specified in the HAT file's link_target attribute.
        The same binary file can be referenced by many HAT files.
        Many HAT packages can exist in the same directory.
        An instance of HATPackage is created by giving HATPackage the file path to the .hat file."""
        self.name = os.path.basename(hat_file_path)
        self.hat_file_path = hat_file_path
        self.hat_file = HATFile.Deserialize(hat_file_path)

        self.link_target = self.hat_file.dependencies.link_target
        self.link_target_path = os.path.join(
            os.path.split(self.hat_file_path)[0], self.hat_file.dependencies.link_target
        )

        self.functions = self.hat_file.functions

    def get_functions(self):
        return self.hat_file.functions

    def get_functions_for_target(self, os: str, arch: str, required_extensions: list = []):
        all_functions = self.get_functions()

        def matches_target(hat_function):
            hat_file = hat_function.hat_file
            if hat_file.target.required.os != os or hat_file.target.required.cpu.architecture != arch:
                return False
            for required_ext in required_extensions:
                if required_ext not in hat_file.target.required.cpu.extensions:
                    return False
            return True

        return list(filter(matches_target, all_functions))


class AttributeDict(OrderedDict):
    """ Dictionary that allows entries to be accessed like attributes
    """
    __getattr__ = OrderedDict.__getitem__

    @property
    def names(self):
        return list(self.keys())

    def __getitem__(self, key):
        for k, v in self.items():
            if k.startswith(key):
                return v
        return OrderedDict.__getitem__(key)


def _make_cpu_func(shared_lib: ctypes.CDLL, function_name: str, arg_infos: List[Parameter]):
    arg_infos = [ArgInfo(d) for d in arg_infos]
    fn = shared_lib[function_name]

    def f(*args):
        # verify that the (numpy) input args match the description in
        # the hat file
        verify_args(args, arg_infos, function_name)

        # prepare the args to the hat package
        hat_args = [arg.ctypes.data_as(arg_info.ctypes_pointer_type) for arg, arg_info in zip(args, arg_infos)]

        # call the function in the hat package
        fn(*hat_args)

    return f


def _make_device_func(func_runtime: str, hat_dir_path: str, func: Function):
    if func_runtime == "CUDA":
        from . import cuda_loader
        return cuda_loader.create_loader_for_device_function(func, hat_dir_path)
    elif func_runtime == "ROCM":
        from . import rocm_loader
        return rocm_loader.create_loader_for_device_function(func, hat_dir_path)


def _load_pkg_binary_module(hat_pkg: HATPackage):
    shared_lib = None
    if os.path.isfile(hat_pkg.link_target_path):

        supported_extensions = [".dll", ".so"]
        _, extension = os.path.splitext(hat_pkg.link_target_path)

        if extension and extension not in supported_extensions:
            # TODO: Should this be an error? Maybe just move on to the
            # device function section?
            raise RuntimeError(f"Unsupported HAT library extension: {extension}")

        hat_binary_path = os.path.abspath(hat_pkg.link_target_path)

        # load the hat_library:
        hat_library = ctypes.cdll.LoadLibrary(hat_binary_path) if extension else None
        shared_lib = hat_library

    return shared_lib


def hat_package_to_func_dict(hat_pkg: HATPackage) -> AttributeDict:

    try:
        try:
            from . import cuda_loader
        except ModuleNotFoundError:
            import cuda_loader
    except:
        CUDA_AVAILABLE = False
    else:
        CUDA_AVAILABLE = True

    try:
        try:
            from . import rocm_loader
        except ModuleNotFoundError:
            import rocm_loader
    except:
        ROCM_AVAILABLE = False
    else:
        ROCM_AVAILABLE = True

    NOTIFY_ABOUT_CUDA = not CUDA_AVAILABLE
    NOTIFY_ABOUT_ROCM = not ROCM_AVAILABLE

    # check that the HAT library has a supported file extension
    func_dict = AttributeDict()
    shared_lib = _load_pkg_binary_module(hat_pkg)
    hat_dir_path, _ = os.path.split(hat_pkg.hat_file_path)

    for func_name, func_desc in hat_pkg.hat_file.function_map.items():

        launches = func_desc.launches
        if not launches and shared_lib:

            func_dict[func_name] = _make_cpu_func(shared_lib, func_desc.name, func_desc.arguments)
        else:
            device_func = hat_pkg.hat_file.device_function_map.get(launches)

            func_runtime = func_desc.runtime
            if not device_func:
                raise RuntimeError(f"Couldn't find device function for loader: " + launches)
            if not func_runtime:
                raise RuntimeError(f"Couldn't find runtime for loader: " + launches)

            # TODO: Generalize this concept to work so it's not CUDA/ROCM specific
            if func_runtime == "CUDA" and not CUDA_AVAILABLE:

                # TODO: printing to stdout only makes sense in tool mode
                if NOTIFY_ABOUT_CUDA:
                    print("CUDA functionality not available on this machine. Please install the cuda python modules")
                    NOTIFY_ABOUT_CUDA = False

                continue

            elif func_runtime == "ROCM" and not ROCM_AVAILABLE:

                # TODO: printing to stdout only makes sense in tool mode
                if NOTIFY_ABOUT_ROCM:
                    print("ROCm functionality not available on this machine. Please install ROCm 4.2 or higher")
                    NOTIFY_ABOUT_ROCM = False

                continue

            func_dict[func_name] = _make_device_func(
                func_runtime=func_runtime, hat_dir_path=hat_dir_path, func=device_func
            )

    return func_dict
